import importlib
import os
import pkgutil
import sys
import typing as t
from collections import defaultdict
from pathlib import Path
from typing import OrderedDict

import click
import yaml
from google.protobuf.timestamp_pb2 import Timestamp
from tabulate import tabulate

from clarifai.utils.logging import logger


def from_yaml(filename: str):
    try:
        with open(filename, 'r') as f:
            return yaml.safe_load(f)
    except yaml.YAMLError as e:
        click.echo(f"Error reading YAML file: {e}", err=True)
        return {}


def dump_yaml(data, filename: str):
    try:
        with open(filename, 'w') as f:
            yaml.dump(data, f)
    except Exception as e:
        click.echo(f"Error writing YAML file: {e}", err=True)


# Dynamically find and import all command modules from the cli directory
def load_command_modules():
    package_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cli')

    for _, module_name, _ in pkgutil.iter_modules([package_dir]):
        if module_name not in ['base', '__main__']:  # Skip the base.py and __main__ file itself
            importlib.import_module(f'clarifai.cli.{module_name}')


def display_co_resources(
    response,
    custom_columns={
        'ID': lambda c: c.id,
        'USER_ID': lambda c: c.user_id,
        'DESCRIPTION': lambda c: c.description,
    },
    sort_by_columns=None,
):
    """
    Display compute orchestration resources listing results using rich.

    :param response: Iterable of resource objects to display.
    :param custom_columns: A dictionary mapping column names to extractor functions.
                           Defaults to ID, USER_ID, and DESCRIPTION.
    :param sort_by_columns: Optional list of (column_name, order) tuples specifying sort order.
                            Only column names present in `custom_columns` are considered.
                            Order should be 'asc' or 'desc'.
    :return: None. Prints the formatted table.
    """
    if sort_by_columns:
        for column, order in reversed(sort_by_columns):  # reversed for stable multi-level sort
            if column in custom_columns:
                reverse = order.lower() == 'desc'
                response = sorted(response, key=custom_columns[column], reverse=reverse)

    formatter = TableFormatter(custom_columns)
    print(formatter.format(list(response), fmt="plain"))


class TableFormatter:
    def __init__(self, custom_columns: OrderedDict):
        """
        Initializes the TableFormatter with column headers and custom column mappings.

        :param headers: List of column headers for the table.
        """
        self.custom_columns = custom_columns

    def format(self, objects, fmt='plain'):
        """
        Formats a list of objects into a table with custom columns.

        :param objects: List of objects to format into a table.
        :return: A string representing the table.
        """
        # Prepare the rows by applying the custom column functions to each object
        rows = []
        for obj in objects:
            #   row = [self.custom_columns[header](obj) for header in self.headers]
            row = [f(obj) for f in self.custom_columns.values()]
            rows.append(row)

        # Create the table
        table = tabulate(rows, headers=self.custom_columns.keys(), tablefmt=fmt)
        return table


class AliasedGroup(click.Group):
    def __init__(
        self,
        name: t.Optional[str] = None,
        commands: t.Optional[
            t.Union[t.MutableMapping[str, click.Command], t.Sequence[click.Command]]
        ] = None,
        **attrs: t.Any,
    ) -> None:
        super().__init__(name, commands, **attrs)
        self.alias_map = {}
        self.command_to_aliases = defaultdict(list)

    def add_alias(self, cmd: click.Command, alias: str) -> None:
        self.alias_map[alias] = cmd
        if alias != cmd.name:
            self.command_to_aliases[cmd].append(alias)

    def command(
        self, aliases=None, *args, **kwargs
    ) -> t.Callable[[t.Callable[..., t.Any]], click.Command]:
        cmd_decorator = super().command(*args, **kwargs)
        if aliases is None:
            aliases = []

        def aliased_decorator(f):
            cmd = cmd_decorator(f)
            if cmd.name:
                self.add_alias(cmd, cmd.name)
            for alias in aliases:
                self.add_alias(cmd, alias)
            return cmd

        f = None
        if args and callable(args[0]):
            (f,) = args
        if f is not None:
            return aliased_decorator(f)
        return aliased_decorator

    def group(
        self, aliases=None, *args, **kwargs
    ) -> t.Callable[[t.Callable[..., t.Any]], click.Group]:
        cmd_decorator = super().group(*args, **kwargs)
        if aliases is None:
            aliases = []

        def aliased_decorator(f):
            cmd = cmd_decorator(f)
            if cmd.name:
                self.add_alias(cmd, cmd.name)
            for alias in aliases:
                self.add_alias(cmd, alias)
            return cmd

        f = None
        if args and callable(args[0]):
            (f,) = args
        if f is not None:
            return aliased_decorator(f)
        return aliased_decorator

    def get_command(self, ctx: click.Context, cmd_name: str) -> t.Optional[click.Command]:
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        return self.alias_map.get(cmd_name)

    def format_commands(self, ctx, formatter):
        sub_commands = self.list_commands(ctx)

        rows = []
        for sub_command in sub_commands:
            cmd = self.get_command(ctx, sub_command)
            if cmd is None or getattr(cmd, 'hidden', False):
                continue
            if cmd in self.command_to_aliases:
                aliases = ', '.join(self.command_to_aliases[cmd])
                sub_command = f'{sub_command} ({aliases})'
            cmd_help = cmd.help
            rows.append((sub_command, cmd_help))

        if rows:
            with formatter.section("Commands"):
                formatter.write_dl(rows)


def validate_context(ctx):
    from clarifai.utils.logging import logger

    if ctx.obj == {}:
        logger.error("CLI config file missing. Run `clarifai login` to set up the CLI config.")
        sys.exit(1)


def validate_context_auth(pat: str, user_id: str, api_base: str = None):
    """
    Validate a Personal Access Token (PAT) by making a test API call.

    Args:
        pat (str): The Personal Access Token to validate
        user_id (str): The user ID associated with the token
        api_base (str): The API base URL. Defaults to None (uses default).
    """
    try:
        from clarifai_grpc.grpc.api.status import status_code_pb2

        from clarifai.client.user import User

        logger.info("Validating the Context Credentials...")

        # Create user client for validation
        if api_base:
            user_client = User(user_id=user_id, pat=pat, base_url=api_base)
        else:
            user_client = User(user_id=user_id, pat=pat)

        # Try to get user info as a test API call
        response = user_client.get_user_info()

        if response.status.code == status_code_pb2.SUCCESS:
            logger.info("✅ Context is valid")

    except Exception as e:
        # Check for common authentication errors and provide user-friendly messages
        logger.error("❌ Authentication failed. Please check your token and user ID.")
        raise click.Abort()  # Exit without saving the configuration


def customize_ollama_model(
    model_path, model_name=None, port=None, context_length=None, verbose=False
):
    """Customize the Ollama model name in the cloned template files.
    Args:
     model_path: Path to the cloned model directory
     model_name: The model name to set (e.g., 'llama3.1', 'mistral') - optional
     port: Port for Ollama server - optional
     context_length: Context length for the model - optional
     verbose: Whether to enable verbose logging - optional (defaults to False)

    """
    model_py_path = os.path.join(model_path, "1", "model.py")

    if not os.path.exists(model_py_path):
        logger.warning(f"Model file {model_py_path} not found, skipping model name customization")
        return

    try:
        # Read the model.py file
        with open(model_py_path, 'r') as file:
            content = file.read()
        if model_name:
            # Replace the default model name in the load_model method
            content = content.replace(
                'self.model = os.environ.get("OLLAMA_MODEL_NAME", \'llama3.2\')',
                f'self.model = os.environ.get("OLLAMA_MODEL_NAME", \'{model_name}\')',
            )

        if port:
            # Replace the default port variable in the model.py file
            content = content.replace("PORT = '23333'", f"PORT = '{port}'")

        if context_length:
            # Replace the default context length variable in the model.py file
            content = content.replace(
                "context_length = '8192'", f"context_length = '{context_length}'"
            )

        verbose_str = str(verbose)
        if "VERBOSE_OLLAMA = True" in content:
            content = content.replace("VERBOSE_OLLAMA = True", f"VERBOSE_OLLAMA = {verbose_str}")
        elif "VERBOSE_OLLAMA = False" in content:
            content = content.replace("VERBOSE_OLLAMA = False", f"VERBOSE_OLLAMA = {verbose_str}")

        # Write the modified content back to model.py
        with open(model_py_path, 'w') as file:
            file.write(content)

    except Exception as e:
        logger.error(f"Failed to customize Ollama model name in {model_py_path}: {e}")
        raise


def check_ollama_installed():
    """Check if the Ollama CLI is installed."""
    try:
        import subprocess

        result = subprocess.run(
            ['ollama', '--version'], capture_output=True, text=True, check=False
        )
        if result.returncode == 0:
            return True
        else:
            return False
    except FileNotFoundError:
        return False


def check_lmstudio_installed():
    """Check if the LM Studio CLI is installed."""
    try:
        import subprocess

        result = subprocess.run(['lms', 'version'], capture_output=True, text=True, check=False)
        if result.returncode == 0:
            return True
        else:
            return False
    except FileNotFoundError:
        return False


def _is_package_installed(package_name):
    """Helper function to check if a single package in requirements.txt is installed."""
    import importlib.metadata

    try:
        importlib.metadata.distribution(package_name)
        logger.debug(f"✅ {package_name} - installed")
        return True
    except importlib.metadata.PackageNotFoundError:
        logger.debug(f"❌ {package_name} - not installed")
        return False
    except Exception as e:
        logger.warning(f"Error checking {package_name}: {e}")
        return False


def parse_requirements(model_path: str):
    """Parse requirements.txt in the model directory and return a dictionary of dependencies."""
    from packaging.requirements import Requirement

    requirements_path = Path(model_path) / "requirements.txt"

    if not requirements_path.exists():
        logger.warning(f"requirements.txt not found at {requirements_path}")
        return []

    deps = {}
    for line in requirements_path.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        try:
            req = Requirement(line)
            deps[req.name] = str(req.specifier) if req.specifier else None
        except Exception as e:
            logger.warning(f"⚠️ Could not parse line: {line!r} — {e}")
    return deps


def check_requirements_installed(model_path):
    """Check if all dependencies in requirements.txt are installed."""

    try:
        # Getting package name and version (for logging)
        requirements = parse_requirements(model_path)

        if not requirements:
            logger.info("No dependencies found in requirements.txt")
            return True

        logger.info(f"Checking {len(requirements)} dependencies...")

        missing = [
            full_req
            for package_name, full_req in requirements.items()
            if not _is_package_installed(package_name)
        ]

        if not missing:
            logger.info(f"✅ All {len(requirements)} dependencies are installed!")
            return True

        # Report missing packages
        logger.error(
            f"❌ {len(missing)} of {len(requirements)} required packages are missing in the current environment"
        )
        logger.error("\n".join(f"  - {pkg}" for pkg in missing))
        requirements_path = Path(model_path) / "requirements.txt"
        logger.warning(f"To install: pip install -r {requirements_path}")
        return False

    except Exception as e:
        logger.error(f"Failed to check requirements: {e}")
        return False


def convert_timestamp_to_string(timestamp: Timestamp) -> str:
    """Converts a Timestamp object to a string.

    Args:
        timestamp (Timestamp): The Timestamp object to convert.

    Returns:
        str: The converted string in ISO 8601 format.
    """
    if not timestamp:
        return ""

    datetime_obj = timestamp.ToDatetime()

    return datetime_obj.strftime('%Y-%m-%dT%H:%M:%SZ')


def customize_huggingface_model(model_path, model_name):
    config_path = os.path.join(model_path, 'config.yaml')
    if os.path.exists(config_path):
        with open(config_path, 'r') as f:
            config = yaml.safe_load(f)

        # Update the repo_id in checkpoints section
        if 'checkpoints' not in config:
            config['checkpoints'] = {}
        config['checkpoints']['repo_id'] = model_name

        with open(config_path, 'w') as f:
            yaml.dump(config, f, default_flow_style=False, sort_keys=False)

        logger.info(f"Updated Hugging Face model repo_id to: {model_name}")
    else:
        logger.warning(f"config.yaml not found at {config_path}, skipping model configuration")


def customize_lmstudio_model(model_path, model_name, port, context_length):
    """Customize the LM Studio model name in the cloned template files.
    Args:
     model_path: Path to the cloned model directory
     model_name: The model name to set (e.g., 'qwen/qwen3-4b-thinking-2507') - optional
     port: Port for LM Studio server - optional
     context_length: Context length for the model - optional

    """
    config_path = os.path.join(model_path, 'config.yaml')

    if os.path.exists(config_path):
        with open(config_path, 'r') as f:
            config = yaml.safe_load(f)
        if 'toolkit' not in config or config['toolkit'] is None:
            config['toolkit'] = {}
        if model_name is not None:
            config['toolkit']['model'] = model_name
        if port is not None:
            config['toolkit']['port'] = port
        if context_length is not None:
            config['toolkit']['context_length'] = context_length
        with open(config_path, 'w') as f:
            yaml.dump(config, f, default_flow_style=False, sort_keys=False)
        logger.info(f"Updated LM Studio model configuration in: {config_path}")
    else:
        logger.warning(f"config.yaml not found at {config_path}, skipping model configuration")



from typing import Dict, Generator, Optional

from clarifai_grpc.grpc.api import resources_pb2, service_pb2

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.constants import DEFAULT_BASE
from clarifai.utils.logging import logger


class Module(Lister, BaseClient):
    """Module is a class that provides access to Clarifai API endpoints related to Module information."""

    def __init__(
        self,
        url: Optional[str] = None,
        module_id: Optional[str] = None,
        module_version: Dict[str, str] = {'id': ""},
        base_url: str = DEFAULT_BASE,
        pat: Optional[str] = None,
        token: Optional[str] = None,
        root_certificates_path: Optional[str] = None,
        **kwargs,
    ):
        """Initializes a Module object.

        Args:
            url (Optional[str]): The URL to initialize the module object.
            module_id (Optional[str]): The Module ID to interact with.
            module_version (Dict[str, str]): The Module Version to interact with.
                                           Defaults to {'id': ""} for latest version.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (Optional[str]): A personal access token for authentication. Can be set as env var CLARIFAI_PAT.
            token (Optional[str]): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN.
            root_certificates_path (Optional[str]): Path to the SSL root certificates file, used to establish secure gRPC connections.
            **kwargs: Additional keyword arguments to be passed to the Module.
        """
        if url and module_id:
            raise UserError("You can only specify one of url or module_id.")
        if not url and not module_id:
            raise UserError("You must specify one of url or module_id.")
        if url:
            user_id, app_id, module_id, module_version_id = ClarifaiUrlHelper.split_module_ui_url(
                url
            )
            module_version = {'id': module_version_id}
            kwargs = {'user_id': user_id, 'app_id': app_id}

        self.kwargs = {**kwargs, 'id': module_id, 'module_version': module_version}
        self.module_info = resources_pb2.Module(**self.kwargs)
        self.logger = logger
        BaseClient.__init__(
            self,
            user_id=self.user_id,
            app_id=self.app_id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self)

    def list_versions(
        self, page_no: int = None, per_page: int = None
    ) -> Generator['Module', None, None]:
        """Lists all the module versions for the module.

        Args:
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Moudle: Module objects for versions of the module.

        Example:
            >>> from clarifai.client.module import Module
            >>> module = Module(module_id='module_id', user_id='user_id', app_id='app_id')
            >>> all_Module_versions = list(module.list_versions())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(
            user_app_id=self.user_app_id,
            module_id=self.id,
        )
        all_module_versions_info = self.list_pages_generator(
            self.STUB.ListModuleVersions,
            service_pb2.ListModuleVersionsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for module_version_info in all_module_versions_info:
            module_version_info['id'] = module_version_info['module_version_id']
            del module_version_info['module_version_id']
            yield Module.from_auth_helper(
                self.auth_helper,
                module_id=self.id,
                **dict(self.kwargs, module_version=module_version_info),
            )

    def __getattr__(self, name):
        return getattr(self.module_info, name)

    def __str__(self):
        init_params = [param for param in self.kwargs.keys()]
        attribute_strings = [
            f"{param}={getattr(self.module_info, param)}"
            for param in init_params
            if hasattr(self.module_info, param)
        ]
        return f"Module Details: \n{', '.join(attribute_strings)}\n"



DATASET_UPLOAD_TASKS = [
    "visual_classification",
    "text_classification",
    "visual_detection",
    "visual_segmentation",
    "visual_captioning",
    "multimodal_dataset",
]

TASK_TO_ANNOTATION_TYPE = {
    "visual_classification": {"concepts": "labels"},
    "text_classification": {"concepts": "labels"},
    "visual_captioning": {"concepts": "labels"},
    "visual_detection": {"bboxes": "bboxes"},
    "visual_segmentation": {"polygons": "polygons"},
}

MAX_RETRIES = 2

CONTENT_TYPE = {"json": "application/json", "zip": "application/zip"}



from collections import namedtuple
from urllib.parse import urlparse

# To help with using ClarifaiUrlHelper with defaults as ClarifaiUrlHelper()
auth_obj = namedtuple("auth", ["ui", "base"])


class ClarifaiUrlHelper(object):
    """Lots of helper functionality for dealing with urls around modules."""

    def __init__(
        self,
        auth=None,
        module_manager_imv_id: str = "module_manager_install",
    ):
        """
        Args:
          auth: a ClarifaiAuthHelper object. Pass None to use the values from the current context.
          module_manager_imv_id: the ID of the module manager installed module version.
        """
        self._auth = auth
        self._module_manager_imv_id = module_manager_imv_id
        self._current_context = None
        if self._auth is None:
            self._auth = auth_obj(self.current_ctx.ui, self.current_ctx.api_base)

    @property
    def ui(self):
        return self._auth.ui

    @property
    def base(self):
        return self._auth.base

    @property
    def current_ctx(self):
        if self._current_context is None:
            from clarifai.utils.config import Config

            self._current_context = Config.from_yaml().current
        return self._current_context

    def module_ui_url(
        self,
        user_id: str = None,
        app_id: str = None,
        module_id: str = None,
        module_version_id: str = None,
    ):
        """This is the path to the module in community."""
        if user_id is None:
            user_id = self.current_ctx.user_id
        if app_id is None:
            app_id = self.current_ctx.app_id
        if module_id is None:
            module_id = self.current_ctx.module_id
        if module_version_id is None:
            module_version_id = self.current_ctx.module_version_id

        return "%s/%s/%s/modules/%s/versions/%s" % (
            self.ui,
            user_id,
            app_id,
            module_id,
            module_version_id,
        )

    def module_install_ui_url(
        self, dest_user_id: str = None, dest_app_id: str = None, module_url: str = None
    ):
        """This is a url that allows for installation of the module from the community at 'module_url'
        into the destination app_id of the destination user_id."""
        if dest_user_id is None:
            dest_user_id = self.current_ctx.user_id
        if dest_app_id is None:
            dest_app_id = self.current_ctx.app_id
        if module_url is None:
            raise ValueError("module_url must be provided to install a module.")
        return "%s/%s/%s/installed_module_versions/%s/install?install=%s" % (
            self.ui,
            dest_user_id,
            dest_app_id,
            self._module_manager_imv_id,
            module_url,
        )

    def imv_ui_url(self, dest_user_id: str = None, dest_app_id: str = None, imv_id: str = None):
        """This is the path to the resource in the UI."""
        if dest_user_id is None:
            dest_user_id = self.current_ctx.user_id
        if dest_app_id is None:
            dest_app_id = self.current_ctx.app_id
        if imv_id is None:
            raise ValueError("imv_id must be provided to get the IMV API URL.")
        return "%s/%s/%s/installed_module_versions/%s" % (
            self.ui,
            dest_user_id,
            dest_app_id,
            imv_id,
        )

    def mcp_api_url(
        self, user_id: str = None, app_id: str = None, model_id: str = None, version_id: str = None
    ):
        """We have a special endpoint for MCP hosted models.

        Example:
          https://api.clarifai.com/v2/ext/mcp/v1/users/{user_id}/apps/{app_id}/models/{model_id}/versions/{version_id}

        Args:
          user_id: the author of the resource.
          app_id: the author's app the resource was created in.
          model_id: the resource ID
          version_id: the version of the resource.
        """
        if user_id is None:
            user_id = self.current_ctx.user_id
        if app_id is None:
            app_id = self.current_ctx.app_id
        if model_id is None:
            model_id = self.current_ctx.model_id
        if version_id is None:
            return "%s/v2/ext/mcp/v1/users/%s/apps/%s/models/%s" % (
                self.base,
                user_id,
                app_id,
                model_id,
            )
        return "%s/v2/ext/mcp/v1/users/%s/apps/%s/models/%s/versions/%s" % (
            self.base,
            user_id,
            app_id,
            model_id,
            version_id,
        )

    def openai_api_url(self):
        """We have a special endpoint for openAI compatible models.

        This doesn't include the /chat/completions suffix which the openAI client automatically
        adds.

        It also doesn't incldue the model which you an set as the model arg in an openAI client call
        using the clarifai_url() method below.

        Example:
          https://api.clarifai.com/v2/ext/openai/v1
        """
        return "%s/v2/ext/openai/v1" % self.base

    def api_url(
        self,
        user_id: str = None,
        app_id: str = None,
        resource_type: str = None,
        resource_id: str = None,
        version_id: str = None,
    ):
        """This is the path to the resource in the API.

        Example:
          https://api.clarifai.com/v2/zeiler/app/modules/module1/versions/2
          https://api.clarifai.com/v2/zeiler/app/models/model1/versions/2
          https://api.clarifai.com/v2/zeiler/app/concepts/concept1
          https://api.clarifai.com/v2/zeiler/app/workflows/workflow1
          https://api.clarifai.com/v2/zeiler/app/tasks/task1
          https://api.clarifai.com/v2/zeiler/app/installed_module_versions/module_manager_install

        Args:
          user_id: the author of the resource.
          app_id: the author's app the resource was created in.
          resource_type: the type of resource. One of "modules", "models", "concepts", "inputs", "workflows", "tasks"
          resource_id: the resource ID
        """
        if user_id is None:
            user_id = self.current_ctx.user_id
        if app_id is None:
            app_id = self.current_ctx.app_id
        if resource_id is None:
            raise ValueError("resource_id must be provided to get the API URL.")
        self._validate_resource_type(resource_type)
        if version_id is None:
            return "%s/v2/users/%s/apps/%s/%s/%s" % (
                self.base,
                user_id,
                app_id,
                resource_type,
                resource_id,
            )

        if resource_type in ["concepts", "tasks", "installed_module_versions"]:
            raise ValueError(f"{resource_type} do not have versions.")
        return "%s/v2/users/%s/apps/%s/%s/%s/versions/%s" % (
            self.base,
            user_id,
            app_id,
            resource_type,
            resource_id,
            version_id,
        )

    def _validate_resource_type(self, resource_type):
        if resource_type not in [
            "modules",
            "models",
            "concepts",
            "inputs",
            "workflows",
            "tasks",
            "installed_module_versions",
        ]:
            raise ValueError(
                "resource_type must be one of modules, models, concepts, inputs, workflows, tasks, installed_module_versions but was %s"
                % resource_type
            )

    def clarifai_url(
        self,
        user_id: str = None,
        app_id: str = None,
        resource_type: str = None,
        resource_id: str = None,
        version_id: str = None,
    ):
        """This is the path to the resource in community UI.

        Example:
          https://clarifai.com/zeiler/modules/module1/versions/2
          https://clarifai.com/zeiler/models/model1/versions/2
          https://clarifai.com/zeiler/concepts/concept1
          https://clarifai.com/zeiler/workflows/workflow1
          https://clarifai.com/zeiler/tasks/task1
          https://clarifai.com/zeiler/installed_module_versions/module_manager_install

        Args:
          user_id: the author of the resource.
          app_id: the author's app the resource was created in.
          resource_type: the type of resource. One of "modules", "models", "concepts", "inputs", "workflows", "tasks", "installed_module_versions"
          resource_id: the resource ID
          version_id: the version of the resource.
        """
        if user_id is None:
            user_id = self.current_ctx.user_id
        if app_id is None:
            app_id = self.current_ctx.app_id
        if resource_id is None:
            raise ValueError("resource_id must be provided to get the API URL.")
        self._validate_resource_type(resource_type)
        if version_id is None:
            return "%s/%s/%s/%s/%s" % (self.ui, user_id, app_id, resource_type, resource_id)
        return "%s/%s/%s/%s/%s/versions/%s" % (
            self.ui,
            user_id,
            app_id,
            resource_type,
            resource_id,
            version_id,
        )

    @classmethod
    def split_clarifai_app_url(cls, url: str):
        """
        clarifai.com uses fully qualified urls to resources.
        They are in the format of:
        https://clarifai.com/{user_id}/{app_id}/
        """
        url = url.replace("https://", "", 1).replace("http://", "", 1)
        o = urlparse(url)
        path = o.path
        path = path.lstrip("/")
        parts = path.split("/")
        if len(parts) != 3:
            raise ValueError(
                f"Provided url must have 2 parts after the domain name. The current parts are: {parts}"
            )
        return tuple(parts[1:])

    @classmethod
    def split_clarifai_url(cls, url: str):
        """
        clarifai.com uses fully qualified urls to resources.
        They are in the format of:
        https://clarifai.com/{user_id}/{app_id}/{resource_type}/{resource_id}/{resource_version_type}/{resource_version_id}
        Those last two are optional.

        """
        url = url.replace("https://", "", 1).replace("http://", "", 1)
        o = urlparse(url)
        path = o.path
        path = path.lstrip("/")
        parts = path.split("/")
        if len(parts) != 5 and len(parts) != 7:
            raise ValueError(
                "Provided url must have 4 or 6 parts after the domain name. These are: {user_id}/{app_id}/{resource_type}/{resource_id}/{resource_version_type}/{resource_version_id}"
            )
        user_id, app_id, resource_type, resource_id = parts[1:5]
        if len(parts) == 7:
            resource_version_id = parts[6]
        else:
            resource_version_id = None
        return user_id, app_id, resource_type, resource_id, resource_version_id

    @classmethod
    def split_module_ui_url(cls, install: str):
        """Takes in a path like https://clarifai.com/zeiler/app/modules/module1/versions/2 to split it apart into it's IDs.

        Returns:
          user_id: the author of the module.
          app_id: the author's app the module was created in.
          module_id: the module ID
          module_version_id: the version of the module.
        """
        user_id, app_id, resource_type, resource_id, resource_version_id = cls.split_clarifai_url(
            install
        )

        if resource_type != "modules" or resource_version_id is None:
            raise ValueError(
                "Provided install url must have 6 parts after the domain name. These are {user_id}/{app_id}/modules/{module_id}/versions/{module_version_id}"
            )
        return user_id, app_id, resource_id, resource_version_id



import json
import os
import sys
from typing import Any, Dict

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.base import BaseClient
from clarifai.runners.utils.pipeline_validation import PipelineConfigValidator
from clarifai.utils.logging import logger


class LiteralBlockDumper(yaml.SafeDumper):
    """Custom YAML dumper that uses literal block style for multi-line strings."""

    def represent_str(self, data):
        if '\n' in data:
            return self.represent_scalar('tag:yaml.org,2002:str', data, style='|')
        return self.represent_scalar('tag:yaml.org,2002:str', data)


LiteralBlockDumper.add_representer(str, LiteralBlockDumper.represent_str)


class PipelineBuilder:
    """Pipeline Builder class for managing pipeline upload to Clarifai."""

    def __init__(self, config_path: str):
        """
        Initialize PipelineBuilder.

        :param config_path: Path to the pipeline configuration file
        """
        self._client = None
        self.config_path = os.path.abspath(config_path)
        self.config_dir = os.path.dirname(self.config_path)
        self.config = self._load_config()
        self.validator = PipelineConfigValidator()
        self.validator.validate_config(self.config)

        # Extract pipeline info
        pipeline_config = self.config["pipeline"]
        self.pipeline_id = pipeline_config["id"]
        self.user_id = pipeline_config["user_id"]
        self.app_id = pipeline_config["app_id"]

        # Track uploaded pipeline step versions
        self.uploaded_step_versions = {}

    @property
    def client(self):
        """Get or create the Clarifai client."""
        if self._client is None:
            self._client = BaseClient(user_id=self.user_id, app_id=self.app_id)
        return self._client

    def _load_config(self) -> Dict[str, Any]:
        """Load and return the configuration from config file."""
        try:
            with open(self.config_path, 'r') as file:
                config = yaml.safe_load(file)
            return config
        except Exception as e:
            raise ValueError(f"Error loading config file {self.config_path}: {e}")

    def _save_config(self) -> None:
        """Save the updated configuration back to the file."""
        try:
            with open(self.config_path, 'w', encoding="utf-8") as file:
                yaml.dump(
                    self.config,
                    file,
                    Dumper=LiteralBlockDumper,
                    default_flow_style=False,
                    sort_keys=False,
                )
        except Exception as e:
            raise ValueError(f"Error saving config file {self.config_path}: {e}")

    def upload_pipeline_steps(self) -> bool:
        """Upload all pipeline steps listed in step_directories."""
        pipeline_config = self.config["pipeline"]
        step_directories = pipeline_config.get("step_directories", [])

        if not step_directories:
            logger.info("No pipeline steps to upload (step_directories is empty)")
            return False  # treat this as an error.

        logger.info(f"Uploading {len(step_directories)} pipeline steps...")

        for step_dir in step_directories:
            step_path = os.path.join(self.config_dir, step_dir)

            if not os.path.exists(step_path):
                logger.error(f"Pipeline step directory not found: {step_path}")
                return False

            logger.info(f"Uploading pipeline step from directory: {step_dir}")

            # Create a backup of the original upload function to capture the version
            # We'll need to modify the upload process to capture the version ID
            success, version_id = self._upload_pipeline_step_with_version_capture(step_path)

            if not success:
                logger.error(f"Failed to upload pipeline step from directory: {step_dir}")
                return False

            # Store the version ID for later use
            self.uploaded_step_versions[step_dir] = version_id
            logger.info(
                f"Successfully uploaded pipeline step {step_dir} with version {version_id}"
            )

        return True

    def _upload_pipeline_step_with_version_capture(self, step_path: str) -> tuple[bool, str]:
        """Upload a pipeline step and capture its version ID."""
        try:
            # Use the existing pipeline step builder
            from clarifai.runners.pipeline_steps.pipeline_step_builder import PipelineStepBuilder

            builder = PipelineStepBuilder(step_path)

            # Create dockerfile if needed
            builder.create_dockerfile()

            # Check if step exists
            exists = builder.check_pipeline_step_exists()
            if exists:
                logger.info(
                    f"Pipeline step {builder.pipeline_step_id} already exists, creating new version"
                )
            else:
                logger.info(f"Creating new pipeline step {builder.pipeline_step_id}")

            # Upload the pipeline step version directly without the interactive prompt
            success = builder.upload_pipeline_step_version()

            if success and builder.pipeline_step_version_id:
                return True, builder.pipeline_step_version_id
            else:
                logger.error("Failed to get pipeline step version ID after upload")
                return False, ""

        except Exception as e:
            logger.error(f"Error uploading pipeline step: {e}")
            return False, ""

    def update_config_with_versions(self) -> None:
        """Update the config.yaml with uploaded pipeline step versions."""
        if not self.uploaded_step_versions:
            logger.info("No pipeline step versions to update in config")
            return

        logger.info("Updating config.yaml with pipeline step versions...")

        # Update the orchestration spec
        pipeline_config = self.config["pipeline"]
        orchestration_spec = pipeline_config["orchestration_spec"]
        argo_spec_str = orchestration_spec["argo_orchestration_spec"]
        argo_spec = yaml.safe_load(argo_spec_str)

        # Update templateRef names to include versions
        self._update_template_refs_with_versions(argo_spec)

        # Update the config
        orchestration_spec["argo_orchestration_spec"] = yaml.dump(
            argo_spec, Dumper=LiteralBlockDumper, default_flow_style=False
        )

        # Remove uploaded directories from step_directories
        remaining_dirs = []
        for step_dir in pipeline_config.get("step_directories", []):
            if step_dir not in self.uploaded_step_versions:
                remaining_dirs.append(step_dir)

        pipeline_config["step_directories"] = remaining_dirs

        # Save the updated config
        self._save_config()
        logger.info("Updated config.yaml with pipeline step versions")

    def _update_template_refs_with_versions(self, argo_spec: Dict[str, Any]) -> None:
        """Update templateRef names in Argo spec to include version information."""
        for template in argo_spec["spec"]["templates"]:
            if "steps" in template:
                for step_group in template["steps"]:
                    for step in step_group:
                        if "templateRef" in step:
                            template_ref = step["templateRef"]
                            name = template_ref["name"]
                            # Extract step name
                            parts = name.split('/')

                            # Check if this is a templateRef without version that we uploaded
                            if self.validator.TEMPLATE_REF_WITHOUT_VERSION_PATTERN.match(name):
                                step_name = parts[-1]
                                # The step name should match the directory name or be derivable from it
                                version_id = self.uploaded_step_versions.get(step_name, None)
                                if version_id is not None:
                                    # Update the templateRef to include version
                                    new_name = f"{name}/versions/{version_id}"
                                    template_ref["name"] = new_name
                                    template_ref["template"] = new_name
                                    logger.info(f"Updated templateRef from {name} to {new_name}")
                            elif self.validator.TEMPLATE_REF_WITH_VERSION_PATTERN.match(name):
                                # strip the /versions/{version_id} from the end of name
                                # to get the name like above
                                orig_name = name
                                name = orig_name.rsplit('/versions/', 1)[0]
                                step_name = parts[-3]  # Get the step name from the path

                                # if it already has a version, make sure it matches the uploaded
                                # version
                                version_id = self.uploaded_step_versions.get(step_name, None)
                                if version_id is not None:
                                    # Update the templateRef to include version
                                    new_name = f"{name}/versions/{version_id}"
                                    template_ref["name"] = new_name
                                    template_ref["template"] = new_name
                                    logger.info(
                                        f"Updated templateRef from {orig_name} to {new_name}"
                                    )

    def create_pipeline(self) -> bool:
        """Create the pipeline using PostPipelines RPC."""
        logger.info(f"Creating pipeline {self.pipeline_id}...")

        try:
            # Create pipeline proto
            pipeline = resources_pb2.Pipeline(
                id=self.pipeline_id, user_id=self.user_id, app_id=self.app_id
            )

            # Add orchestration spec
            pipeline_config = self.config["pipeline"]
            orchestration_spec = pipeline_config["orchestration_spec"]
            argo_spec_str = orchestration_spec["argo_orchestration_spec"]

            # Parse the Argo spec to get API version
            argo_spec = yaml.safe_load(argo_spec_str)
            api_version = argo_spec.get("apiVersion", "argoproj.io/v1alpha1")

            # Create pipeline version with orchestration spec
            pipeline_version = resources_pb2.PipelineVersion()
            # Create orchestration spec proto
            orchestration_spec_proto = resources_pb2.OrchestrationSpec()
            # Create Argo orchestration spec proto
            argo_orchestration_spec_proto = resources_pb2.ArgoOrchestrationSpec()
            argo_orchestration_spec_proto.api_version = api_version
            argo_orchestration_spec_proto.spec_json = json.dumps(argo_spec)

            orchestration_spec_proto.argo_orchestration_spec.CopyFrom(
                argo_orchestration_spec_proto
            )
            pipeline_version.orchestration_spec.CopyFrom(orchestration_spec_proto)

            pipeline.pipeline_version.CopyFrom(pipeline_version)

            # Make the RPC call
            response = self.client.STUB.PostPipelines(
                service_pb2.PostPipelinesRequest(
                    user_app_id=self.client.user_app_id, pipelines=[pipeline]
                )
            )

            if response.status.code == status_code_pb2.SUCCESS:
                logger.info(f"Successfully created pipeline {self.pipeline_id}")

                # Log pipeline and version IDs if available in response
                if response.pipelines:
                    created_pipeline = response.pipelines[0]
                    logger.info(f"Pipeline ID: {created_pipeline.id}")
                    if created_pipeline.pipeline_version and created_pipeline.pipeline_version.id:
                        logger.info(f"Pipeline version ID: {created_pipeline.pipeline_version.id}")

                return True
            else:
                logger.error(f"Failed to create pipeline: {response.status.description}")
                logger.error(f"Details: {response.status.details}")
                return False

        except Exception as e:
            logger.error(f"Error creating pipeline: {e}")
            return False


def upload_pipeline(path: str):
    """
    Upload a pipeline with associated pipeline steps to Clarifai.

    :param path: Path to the pipeline configuration file or directory containing config.yaml
    """
    try:
        # Determine if path is a directory or file
        if os.path.isdir(path):
            config_path = os.path.join(path, "config.yaml")
            if not os.path.exists(config_path):
                raise FileNotFoundError(f"config.yaml not found in directory: {path}")
        else:
            config_path = path

        builder = PipelineBuilder(config_path)

        logger.info(f"Starting pipeline upload from config: {config_path}")

        # Step 1: Upload pipeline steps
        if not builder.upload_pipeline_steps():
            logger.error("Failed to upload pipeline steps")
            sys.exit(1)

        # Step 2: Update config with version information
        builder.update_config_with_versions()

        # Step 3: Create the pipeline
        if not builder.create_pipeline():
            logger.error("Failed to create pipeline")
            sys.exit(1)

        logger.info("Pipeline upload completed successfully!")

    except Exception as e:
        logger.error(f"Pipeline upload failed: {e}")
        sys.exit(1)



# -*- coding: utf-8 -*-
import json
import time

import requests  # noqa
from google.protobuf.json_format import MessageToDict

from clarifai.versions import CLIENT_VERSION, OS_VER, PYTHON_VERSION


class TokenError(Exception):
    pass


class ApiError(Exception):
    """API Server error"""

    def __init__(
        self, resource: str, params: dict, method: str, response: requests.Response = None
    ) -> None:
        self.resource = resource
        self.params = params
        self.method = method
        self.response = response

        self.error_code = 'N/A'
        self.error_desc = 'N/A'
        self.error_details = 'N/A'
        response_json = 'N/A'

        if response is not None:
            response_json_dict = MessageToDict(response)

            self.error_code = response_json_dict.get('status', {}).get('code', None)
            self.error_desc = response_json_dict.get('status', {}).get('description', None)
            self.error_details = response_json_dict.get('status', {}).get('details', None)
            response_json = json.dumps(response_json_dict['status'], indent=2)

        current_ts_str = str(time.time())

        msg = """%(method)s %(resource)s FAILED(%(time_ts)s).  error_code: %(error_code)s, error_description: %(error_desc)s, error_details: %(error_details)s
 >> Python client %(client_version)s with Python %(python_version)s on %(os_version)s
 >> %(method)s %(resource)s
 >> REQUEST(%(time_ts)s) %(request)s
 >> RESPONSE(%(time_ts)s) %(response)s""" % {
            'method': method,
            'resource': resource,
            'error_code': self.error_code,
            'error_desc': self.error_desc,
            'error_details': self.error_details,
            'request': json.dumps(params, indent=2),
            'response': response_json,
            'time_ts': current_ts_str,
            'client_version': CLIENT_VERSION,
            'python_version': PYTHON_VERSION,
            'os_version': OS_VER,
        }

        super(ApiError, self).__init__(msg)


class ApiClientError(Exception):
    """API Client Error"""


class UserError(Exception):
    """User Error"""


class AuthError(Exception):
    """Raised when a client has missing or invalid authentication."""


def _base_url(url: str) -> str:
    """
    Extracts the base URL from the url, which is everything before the 4th slash character.
    https://www.clarifai.com/v2/models/1/output -> https://www.clarifai.com/v2/
    """
    try:
        return url[: _find_nth(url, '/', 4) + 1]
    except Exception:
        return ''


def _find_nth(haystack: str, needle: str, n: int) -> int:
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start + len(needle))
        n -= 1
    return start



import os


class ClarifaiStreamlitCSS(object):
    """ClarifaiStreamlitCSS helps get a consistent style by default for Clarifai provided
    streamlit apps.
    """

    @classmethod
    def insert_default_css(cls, st):
        """Inserts the default style provided in style.css in this folder into the streamlit page

        Example:
          ClarifaiStreamlitCSS.insert_default_css()

        Note:
          This must be placed in both the app.py AND all the pages/*.py files to get the custom styles.
        """
        file_name = os.path.join(os.path.dirname(__file__), "style.css")
        cls.insert_css_file(file_name, st)

    @classmethod
    def insert_css_file(cls, css_file, st):
        """Open the full filename to the css file and insert it's contents the style of the page."""
        with open(css_file) as f:
            st.markdown(f'<style>{f.read()}</style>', unsafe_allow_html=True)

    @classmethod
    def buttonlink(cls, st, label, link, target="_parent", style=""):
        """
        This is a streamlit button that will link to another page (or _self if target is _self).
        It is styled to look like the other stButton>button buttons that are created with st.button().

        You must insert_default_css(st) before using on a page.

        Example:
          ClarifaiStreamlitCSS.insert_default_css(st)
          cols = st.columns(4)
          ClarifaiStreamlitCSS.buttonlink(cols[3], "Button", "https://clarifai.com", "_blank")

        Args:
          st: the streamlit package.
          label: the text string to display in the button.
          link: the url to link the button to.
          target: to open the link in same page (_parent) or new tab (_blank).
          style: additional style to apply to the button link.
            Ex: "background-color: rgb(45, 164, 78); color:white;" makes the button green background with white text.
        """
        astyle = ""
        if style:
            astyle = f'style="{style}"'

        st.markdown(
            f'''
    <div class="stButton">
      <a href="{link}" target="{target}" {astyle}>{label}</a>
    </div>
    ''',
            unsafe_allow_html=True,
        )



import shutil

import click

from clarifai.cli.base import cli
from clarifai.utils.cli import AliasedGroup, display_co_resources, from_yaml, validate_context


@cli.group(
    ['deployment', 'dp'],
    cls=AliasedGroup,
    context_settings={'max_content_width': shutil.get_terminal_size().columns - 10},
)
def deployment():
    """Manage Deployments: create, delete, list"""


@deployment.command(['c'])
@click.argument('nodepool_id')
@click.argument('deployment_id')
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the deployment config file.',
)
@click.pass_context
def create(ctx, nodepool_id, deployment_id, config):
    """Create a new Deployment with the given config file."""

    from clarifai.client.nodepool import Nodepool

    validate_context(ctx)
    if not nodepool_id:
        deployment_config = from_yaml(config)
        nodepool_id = deployment_config['deployment']['nodepools'][0]['id']

    nodepool = Nodepool(
        nodepool_id=nodepool_id,
        user_id=ctx.obj.current.user_id,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base,
    )
    if deployment_id:
        nodepool.create_deployment(config, deployment_id=deployment_id)
    else:
        nodepool.create_deployment(config)


@deployment.command(['ls'])
@click.argument('nodepool_id', default="")
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.pass_context
def list(ctx, nodepool_id, page_no, per_page):
    """List all deployments for the nodepool."""
    from clarifai.client.compute_cluster import ComputeCluster
    from clarifai.client.nodepool import Nodepool
    from clarifai.client.user import User

    validate_context(ctx)
    if nodepool_id:
        nodepool = Nodepool(
            nodepool_id=nodepool_id,
            user_id=ctx.obj.current.user_id,
            pat=ctx.obj.current.pat,
            base_url=ctx.obj.current.api_base,
        )
        response = nodepool.list_deployments(page_no=page_no, per_page=per_page)
    else:
        user = User(
            user_id=ctx.obj.current.user_id,
            pat=ctx.obj.current.pat,
            base_url=ctx.obj.current.api_base,
        )
        ccs = user.list_compute_clusters(page_no, per_page)
        nps = []
        for cc in ccs:
            compute_cluster = ComputeCluster(
                compute_cluster_id=cc.id,
                user_id=ctx.obj.current.user_id,
                pat=ctx.obj.current.pat,
                base_url=ctx.obj.current.api_base,
            )
            nps.extend([i for i in compute_cluster.list_nodepools(page_no, per_page)])
        response = []
        for np in nps:
            nodepool = Nodepool(
                nodepool_id=np.id,
                user_id=ctx.obj.current.user_id,
                pat=ctx.obj.current.pat,
                base_url=ctx.obj.current.api_base,
            )
            response.extend(
                [i for i in nodepool.list_deployments(page_no=page_no, per_page=per_page)]
            )

    display_co_resources(
        response,
        custom_columns={
            'ID': lambda c: c.id,
            'USER_ID': lambda c: c.user_id,
            'COMPUTE_CLUSTER_ID': lambda c: c.nodepools[0].compute_cluster.id,
            'NODEPOOL_ID': lambda c: c.nodepools[0].id,
            'MODEL_USER_ID': lambda c: c.worker.model.user_id,
            'MODEL_APP_ID': lambda c: c.worker.model.app_id,
            'MODEL_ID': lambda c: c.worker.model.id,
            'MODEL_VERSION_ID': lambda c: c.worker.model.model_version.id,
            'DESCRIPTION': lambda c: c.description,
        },
    )


@deployment.command(['rm'])
@click.argument('nodepool_id')
@click.argument('deployment_id')
@click.pass_context
def delete(ctx, nodepool_id, deployment_id):
    """Deletes a deployment for the nodepool."""
    from clarifai.client.nodepool import Nodepool

    validate_context(ctx)
    nodepool = Nodepool(
        nodepool_id=nodepool_id,
        user_id=ctx.obj.current.user_id,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base,
    )
    nodepool.delete_deployments([deployment_id])



import os
import urllib.request
from typing import Any, Dict, Tuple

import grpc
from clarifai_grpc.channel.clarifai_channel import ClarifaiChannel
from clarifai_grpc.grpc.api import resources_pb2, service_pb2_grpc

from clarifai import __version__
from clarifai.utils.constants import (
    CLARIFAI_PAT_ENV_VAR,
    CLARIFAI_SESSION_TOKEN_ENV_VAR,
    DEFAULT_BASE,
    DEFAULT_UI,
)

REQUEST_ID_PREFIX_HEADER = "x-clarifai-request-id-prefix"
REQUEST_ID_PREFIX = f"sdk-python-{__version__}"

# Map from base domain to True / False for whether the base has https or http.
# This is filled in get_stub() if it's not in there already.
base_https_cache = {}
ui_https_cache = {}


def clear_cache() -> None:
    """Clears the cache."""
    base_https_cache.clear()
    ui_https_cache.clear()


def https_cache(cache: dict, url: str) -> str:
    """This is a helper function to cache whether a url is http or https."""
    HTTPS = True
    HTTP = False

    # If http or https is provided, we trust that it is correct.
    # Note: this always stores the url without http:// or https://
    if url.startswith("https://"):
        url = url.replace("https://", "")
        cache[url] = HTTPS
    elif url.startswith("http://"):
        url = url.replace("http://", "")
        cache[url] = HTTP
    elif url.endswith(":443"):
        # If it ends with :443 then we know it's https.
        # trim it off the right
        url = url[:-4]
        cache[url] = HTTPS
    elif url.find('.clarifai.com') >= 0:
        # We know our endpoints are https.
        cache[url] = HTTPS
    elif url not in cache:
        # need to test ones that we don't have in the cache yet.
        try:  # make request to https endpoint.
            urllib.request.urlopen("https://%s/v2/auth/methods" % url, timeout=5)
            cache[url] = HTTPS  # cache it.
        except Exception as e:
            if "SSL" in str(e):  # if ssl error then we know it's http.
                cache[url] = HTTP
                # For http urls we need host:port format.
                if ":" not in url:
                    raise Exception(
                        "When providing an insecure url it must have both host:port format"
                    )
            else:
                raise Exception(
                    "Could not get a valid response from url: %s, is the API running there?" % url
                ) from e
    return url


class ClarifaiAuthHelper:
    def __init__(
        self,
        user_id: str = "",
        app_id: str = "",
        pat: str = "",
        token: str = "",
        base: str = DEFAULT_BASE,
        ui: str = DEFAULT_UI,
        root_certificates_path: str = None,
        validate: bool = True,
    ):
        """
            A helper to get the authorization information needed to make API calls with the grpc
            client to a specific app using a personal access token.

            There are classmethods to make this object easily from either query_params provided by streamlit or from env vars.

            Note: only one of personal access token (pat) or a session token (token) can be provided.
            Always use PATs in your code and never session tokens, those are only provided internal UI code.

        Args:
          user_id: a user id who owns the resource you want to make calls to.
          app_id: an app id for the application that owns the resource you want to interact with
          pat: a personal access token.
          token: a session token (internal use only, always use a PAT).
          base: a url to the API endpoint to hit. Examples include api.clarifai.com,
        https://api.clarifai.com (default), https://host:port, http://host:port, host:port (will be treated as http, not https). It's highly recommended to include the http:// or https:// otherwise we need to check the endpoint to determine if it has SSL during this __init__
          ui: a url to the UI. Examples include clarifai.com,
        https://clarifai.com (default), https://host:port, http://host:port, host:port (will be treated as http, not https). It's highly recommended to include the http:// or https:// otherwise we need to check the endpoint to determine if it has SSL during this __init__
          root_certificates_path: path to the root certificates file. This is only used for grpc secure channels.
          validate: whether to validate the inputs. This is useful for overriding vars then validating
        """

        self.user_id = user_id
        self.app_id = app_id
        self._pat = pat
        self._token = token
        self._root_certificates_path = root_certificates_path

        self.set_base(base)
        self.set_ui(ui)
        if validate:
            self.validate()

    def validate(self):
        if self.user_id == "":
            raise Exception(
                "Need 'user_id' to not be empty in the query params or user CLARIFAI_USER_ID env var"
            )
        if self._pat != "" and self._token != "":
            raise Exception(
                "A personal access token OR a session token need to be provided, but you cannot provide both."
            )
        elif self._pat == "" and self._token == "":
            raise Exception(
                "Need 'pat' or 'token' in the query params or use one of the CLARIFAI_PAT or CLARIFAI_SESSION_TOKEN env vars"
            )
        if (self._root_certificates_path) and (not os.path.exists(self._root_certificates_path)):
            raise Exception(
                "Root certificates path %s does not exist" % self._root_certificates_path
            )

    @classmethod
    def from_streamlit(cls, st: Any) -> "ClarifaiAuthHelper":
        """This is a convenient method to check the environment variables first to see if there are
        required variables for auth, then override them with any additional query parameters that may
        have been passed in.

        Note: if a .streamlit/secrets.toml is present then st.secrets will auto populate the
        corresponding environment variables and we will pick them up from there, OVERWRITING whatever
        matching env var that may already be present.

            Args:
              st: the streamlit package typically as: 'import streamlit as st'
            Returns:
              auth: this class instantiated
        """
        # start with the env vars (potentially loaded from secrets.toml)
        # Don't validate yet as we'll layer on the query params next.
        auth = ClarifaiAuthHelper.from_env(validate=False)

        # Then add in the query params.
        try:
            if st.query_params:
                auth.add_streamlit_query_params(st.query_params)
            else:
                auth.add_streamlit_query_params(st.session_state)
        except Exception as e:
            st.error(e)
            st.stop()
            raise e

        # Then validate.
        try:
            auth.validate()
        except Exception as e:
            st.error(e)
            st.stop()
            raise e

        return auth

    @classmethod
    def from_streamlit_query_params(cls, query_params: Any = "") -> "ClarifaiAuthHelper":
        """Initialize from streamlit queryparams. The following things will be looked for:
        user_id: as 'user_id' in query_params
        app_id: as 'app_id' in query_params
        one of:
          token: as 'token' in query_params
          pat: as 'pat' in query_params
        optionally:
          base: as 'base' in query_params.
          ui: as 'ui' in query_params.

        """

        # Setup an empty one (not from env).
        auth = ClarifaiAuthHelper("", "", "", "", validate=False)

        # Then add in the query params.
        auth.add_streamlit_query_params(query_params)

        # Then validate.
        auth.validate()

        return auth

    def add_streamlit_query_params(self, query_params: Any = "") -> "ClarifaiAuthHelper":
        """Initialize from streamlit queryparams. The following things will be looked for:
              user_id: as 'user_id' in query_params
              app_id: as 'app_id' in query_params
              one of:
                token: as 'token' in query_params
                pat: as 'pat' in query_params
              optionally:
                base: as 'base' in query_params.

        Args:
          query_params: the streamlit.query_params response or streamlit.session_state.
        """

        if query_params == "":  # empty response from streamlit
            query_params = {}
        if "user_id" in query_params:
            self.user_id = query_params["user_id"]
        if "app_id" in query_params:
            self.app_id = query_params["app_id"]
        if "token" in query_params:
            self._token = query_params["token"]
        if "pat" in query_params:
            self._pat = query_params["pat"]
        if "base" in query_params:
            self.set_base(query_params["base"])
        if "ui" in query_params:
            self.set_ui(query_params["ui"])
        if "root_certificates_path" in query_params:
            self._root_certificates_path = query_params["root_certificates_path"]

    @classmethod
    def from_env(cls, validate: bool = True) -> "ClarifaiAuthHelper":
        """Will look for the following env vars:
        user_id: CLARIFAI_USER_ID env var.
        app_id: CLARIFAI_APP_ID env var.
        one of:
          token: CLARIFAI_SESSION_TOKEN env var.
          pat: CLARIFAI_PAT env var.
        base: CLARIFAI_API_BASE env var.
        root_certificates_path: CLARIFAI_ROOT_CERTIFICATES_PATH env var.
        """
        user_id = os.environ.get("CLARIFAI_USER_ID", "")
        app_id = os.environ.get("CLARIFAI_APP_ID", "")
        token = os.environ.get(CLARIFAI_SESSION_TOKEN_ENV_VAR, "")
        pat = os.environ.get(CLARIFAI_PAT_ENV_VAR, "")
        base = os.environ.get("CLARIFAI_API_BASE", DEFAULT_BASE)
        ui = os.environ.get("CLARIFAI_UI", DEFAULT_UI)
        root_certificates_path = os.environ.get("CLARIFAI_ROOT_CERTIFICATES_PATH", None)
        return cls(user_id, app_id, pat, token, base, ui, root_certificates_path, validate)

    def get_user_app_id_proto(
        self,
        user_id: str = None,
        app_id: str = None,
    ) -> resources_pb2.UserAppIDSet:
        """Get the gRPC metadata that contains either the session token or the PAT to use.

        Args:
          user_id: optional user_id to override the default
          app_id: optional app_id to override the default.

        Returns:
          metadata: the metadata need to send with all grpc API calls in the API client.
        """
        user_id = self.user_id if user_id is None else user_id
        app_id = self.app_id if app_id is None else app_id
        return resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)

    @property
    def metadata(self):
        """Get the gRPC metadata that contains either the session token or the PAT to use.

        Returns:
          metadata: the metadata need to send with all grpc API calls in the API client.
        """
        if self._pat != "":
            return (
                ("authorization", "Key %s" % self._pat),
                (REQUEST_ID_PREFIX_HEADER, REQUEST_ID_PREFIX),
            )
        elif self._token != "":
            return (
                ("x-clarifai-session-token", self._token),
                (REQUEST_ID_PREFIX_HEADER, REQUEST_ID_PREFIX),
            )
        else:
            raise Exception(
                "'token' or 'pat' needed to be provided in the query params or env vars."
            )

    def get_stub(self) -> service_pb2_grpc.V2Stub:
        stub, channel = self.get_stub_and_channel()
        return stub

    def get_stub_and_channel(self) -> Tuple[service_pb2_grpc.V2Stub, grpc.Channel]:
        """Get the API gRPC stub and channel based on the API endpoint base.

        Returns:
          stub: The service_pb2_grpc.V2Stub stub for the API.
        """
        if self._base not in base_https_cache:
            raise Exception("Cannot determine if base %s is https" % self._base)

        https = base_https_cache[self._base]
        if https:
            channel = ClarifaiChannel.get_grpc_channel(
                base=self._base, root_certificates_path=self._root_certificates_path
            )
        else:
            if self._base.find(":") >= 0:
                host, port = self._base.split(":")
            else:
                host = self._base
                port = 80
            channel = ClarifaiChannel.get_insecure_grpc_channel(base=host, port=port)
        stub = service_pb2_grpc.V2Stub(channel)
        return stub, channel

    def get_async_stub(self) -> service_pb2_grpc.V2Stub:
        """Get the API gRPC async stub using the right channel based on the API endpoint base.
        Returns:
        stub: The service_pb2_grpc.V2Stub stub for the API.
        """
        if self._base not in base_https_cache:
            raise Exception("Cannot determine if base %s is https" % self._base)

        https = base_https_cache[self._base]
        if https:
            channel = ClarifaiChannel.get_aio_grpc_channel(
                base=self._base, root_certificates_path=self._root_certificates_path
            )
        else:
            if self._base.find(":") >= 0:
                host, port = self._base.split(":")
            else:
                host = self._base
                port = 80
            channel = ClarifaiChannel.get_aio_insecure_grpc_channel(base=host, port=port)
        stub = service_pb2_grpc.V2Stub(channel)
        return stub

    @property
    def ui(self) -> str:
        """Return the domain for the UI."""
        if self._ui not in ui_https_cache:
            raise Exception("Cannot determine if ui %s is https" % self._ui)
        https = ui_https_cache[self._ui]
        if https:
            if not self._ui.startswith("https://"):
                return "https://" + self._ui
            return self._ui
        if not self._ui.startswith("http://"):
            return "http://" + self._ui
        return self._ui

    def set_base(self, base: str):
        """Set the base domain for the API."""
        base = DEFAULT_BASE if base is None else base
        self._base = https_cache(base_https_cache, base)

    def set_ui(self, ui: str):
        """Set the domain for the UI."""
        self._ui = https_cache(ui_https_cache, ui)

    @property
    def base(self) -> str:
        """Return the base domain for the API."""
        if self._base not in base_https_cache:
            raise Exception("Cannot determine if base %s is https" % self._base)
        https = base_https_cache[self._base]
        if https:
            if not self._base.startswith("https://"):
                return "https://" + self._base
            return self._base
        if not self._base.startswith("http://"):
            return "http://" + self._base
        return self._base

    @property
    def pat(self) -> str:
        """Return the personal access token."""
        return self._pat

    def __str__(self):
        return "ClarifaiAuthHelper:\n- base: %s\n- user_id: %s\n- app_id: %s\n" % (
            self._base,
            self.user_id,
            self.app_id,
        )

    @classmethod
    def required_env_vars(cls):
        """Return the list of the required environment variables."""
        return ["CLARIFAI_USER_ID", "CLARIFAI_APP_ID", "CLARIFAI_PAT"]

    @classmethod
    def validate_secrets_dict(cls, toml_dict: Dict[str, Any]):
        """Validate the secrets.toml file has been filled with non-empty values for all the auth
        parameters that are present.

        We don't load the file here so that we don't need the tomli package dependency. You can simply
        do:
        import tomli

        d = tomli.load(open("secrets.toml"))
        ClarifaiAuthHelper.validate_secrets_dict(d)

        """
        # We don't validate the bases because they have sensible defaults
        auth_keys = cls.required_env_vars()

        for k, v in toml_dict.items():
            if k in auth_keys:
                if v == "":
                    raise Exception("'%s' in secrets.toml cannot be empty" % k)
        # for all the keys that are not present, they have a non empty value.
        return True



import time
import uuid
from typing import Dict, List

from clarifai.runners.utils.data_types import Audio, Image, Video
from clarifai.runners.utils.data_utils import process_audio, process_image, process_video


def generate_id():
    return f"chatcmpl-{uuid.uuid4().hex}"


def _format_non_streaming_response(
    generated_text,
    model="custom-model",
    id=None,
    created=None,
    prompt_tokens=None,
    completion_tokens=None,
    finish_reason="stop",
):
    if id is None:
        id = generate_id()
    if created is None:
        created = int(time.time())

    response = {
        "id": id,
        "object": "chat.completion",
        "created": created,
        "model": model,
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": generated_text,
                },
                "finish_reason": finish_reason,
                "logprobs": None,
            }
        ],
    }

    if prompt_tokens is not None and completion_tokens is not None:
        response["usage"] = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": prompt_tokens + completion_tokens,
        }

    return response


def _format_streaming_response(
    generated_chunks,
    model="custom-model",
    id=None,
    created=None,
    finish_reason="stop",
):
    if id is None:
        id = generate_id()
    if created is None:
        created = int(time.time())

    for chunk in generated_chunks:
        yield {
            "id": id,
            "object": "chat.completion.chunk",
            "created": created,
            "model": model,
            "choices": [
                {
                    "index": 0,
                    "delta": {
                        "content": chunk,
                    },
                    "finish_reason": None,
                    "logprobs": None,
                }
            ],
        }

    # Final chunk indicating completion
    yield {
        "id": id,
        "object": "chat.completion.chunk",
        "created": created,
        "model": model,
        "choices": [
            {
                "index": 0,
                "delta": {},
                "finish_reason": finish_reason,
                "logprobs": None,
            }
        ],
    }


def openai_response(
    generated_text,
    model="custom-model",
    id=None,
    created=None,
    prompt_tokens=None,
    completion_tokens=None,
    finish_reason="stop",
    stream=True,
):
    if stream:
        return _format_streaming_response(generated_text, model, id, created, finish_reason)
    else:
        return _format_non_streaming_response(
            generated_text, model, id, created, prompt_tokens, completion_tokens, finish_reason
        )


def openai_to_hf_messages(openai_messages):
    """
    Converts OpenAI-style chat messages into a format compatible with Hugging Face's
    `tokenizer.apply_chat_template()` function, supporting all modalities (text, images, etc.).

    Args:
        openai_messages (list): List of OpenAI-style messages, where each message is a dict with
                                'role' (str) and 'content' (str or list of parts).

    Returns:
        list: Hugging Face-compatible messages. Each message is a dict with 'role' and 'content'.
              Content is a string (text-only) or a list of parts (multimodal).
    """
    hf_messages = []
    for msg in openai_messages:
        role = msg['role']
        content = msg['content']

        if isinstance(content, list):
            # Handle multimodal content (e.g., text + images)
            converted_content = []
            for part in content:
                if part['type'] == 'text':
                    converted_content.append({'type': 'text', 'text': part['text']})
                elif part['type'] == 'image_url':
                    # Handle image (extract base64 or URL)
                    image_url = part["image_url"]["url"]
                    if image_url.startswith("data:image"):
                        # Base64-encoded image
                        b64_img = image_url.split(",")[1]
                        converted_content.append({'type': 'image', 'base64': b64_img})
                    else:
                        # URL (model must handle downloads)
                        converted_content.append({'type': 'image', 'url': image_url})
                elif part['type'] == 'video_url':
                    video_url = part["video_url"]["url"]
                    if video_url.startswith("data:video"):
                        raise ValueError("Base64 video data is not supported in HF format.")
                    else:
                        # URL (model must handle downloads)
                        converted_content.append({'type': 'video', 'url': video_url})
                else:
                    raise ValueError(f"Unsupported content type: {part['type']} for conversion.")
            hf_content = converted_content
        else:
            # Text-only content (string)
            hf_content = content

        hf_messages.append({'role': role, 'content': hf_content})

    return hf_messages


def build_openai_messages(
    prompt: str = None,
    image: Image = None,
    images: List[Image] = None,
    audio: Audio = None,
    audios: List[Audio] = None,
    video: Video = None,
    videos: List[Video] = None,
    messages: List[Dict] = None,
) -> List[Dict]:
    """
    Construct OpenAI-compatible messages from input components.
      Args:
          prompt (str): The prompt text.
          image (Image): Clarifai Image object.
          images (List[Image]): List of Clarifai Image objects.
          audio (Audio): Clarifai Audio object.
          audios (List[Audio]): List of Clarifai Audio objects.
          video (Video): Clarifai Video object.
          videos (List[Video]): List of Clarifai Video objects.
          messages (List[Dict]): List of chat messages.
      Returns:
          List[Dict]: Formatted chat messages.
    """

    openai_messages = []
    # Add previous conversation history
    if messages and is_openai_chat_format(messages):
        openai_messages.extend(messages)

    content = []
    if prompt.strip():
        # Build content array for current message
        content.append({'type': 'text', 'text': prompt})
    # Add single image if present
    if image:
        content.append(process_image(image))
    # Add multiple images if present
    if images:
        for img in images:
            content.append(process_image(img))
    # Add single audio if present
    if audio:
        content.append(process_audio(audio))
    # Add multiple audios if present
    if audios:
        for audio in audios:
            content.append(process_audio(audio))
    # Add single video if present
    if video:
        content.append(process_video(video))
    # Add multiple videos if present
    if videos:
        for video in videos:
            content.append(process_video(video))

    if content:
        # Append complete user message
        openai_messages.append({'role': 'user', 'content': content})

    return openai_messages


def is_openai_chat_format(messages):
    """
    Verify if the given argument follows the OpenAI chat messages format.

    Args:
        messages (list): A list of dictionaries representing chat messages.

    Returns:
        bool: True if valid, False otherwise.
    """
    if not isinstance(messages, list):
        return False

    valid_roles = {"system", "user", "assistant", "function"}

    for msg in messages:
        if not isinstance(msg, dict):
            return False
        if "role" not in msg or "content" not in msg:
            return False
        if msg["role"] not in valid_roles:
            return False

        content = msg["content"]

        # Content should be either a string (text message) or a multimodal list
        if isinstance(content, str):
            continue  # Valid text message

        elif isinstance(content, list):
            for item in content:
                if not isinstance(item, dict):
                    return False
    return True



import asyncio
import inspect
import time
from typing import Any, Dict, Iterator, List

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.auth.register import V2Stub
from clarifai.constants.model import MAX_MODEL_PREDICT_INPUTS
from clarifai.errors import UserError
from clarifai.runners.utils import code_script, method_signatures
from clarifai.runners.utils.method_signatures import (
    CompatibilitySerializer,
    deserialize,
    get_stream_from_signature,
    serialize,
    signatures_from_json,
)
from clarifai.runners.utils.openai_convertor import is_openai_chat_format
from clarifai.utils.logging import logger
from clarifai.utils.misc import BackoffIterator, status_is_retryable


def is_async_context():
    """Check if code is running in an async context."""
    try:
        asyncio.get_running_loop()
        import sys

        # In Jupyter, to check if we're actually in an async cell. Becaue by default jupyter considers it as async.
        if 'ipykernel' in sys.modules:
            return False
        return True
    except RuntimeError:
        return False


class ModelClient:
    '''
    Client for calling model predict, generate, and stream methods.
    '''

    def __init__(
        self,
        stub,
        async_stub: V2Stub = None,
        request_template: service_pb2.PostModelOutputsRequest = None,
    ):
        '''
        Initialize the model client.

        Args:
            stub: The gRPC stub for the model.
            request_template: The template for the request to send to the model, including
            common fields like model_id, model_version, cluster, etc.
        '''
        self.STUB = stub
        self.async_stub = async_stub
        self.request_template = request_template or service_pb2.PostModelOutputsRequest()
        self._method_signatures = None
        self._defined = False

    def fetch(self):
        '''
        Fetch function signature definitions from the model and define the functions in the client
        '''
        if self._defined:
            return
        try:
            self._fetch_signatures()
            self._define_functions()
        finally:
            self._defined = True

    def __getattr__(self, name):
        if not self._defined:
            self.fetch()
        try:
            return self.__getattribute__(name)
        except AttributeError as e:
            # Provide helpful error message with available methods
            available_methods = []
            if self._method_signatures:
                available_methods = list(self._method_signatures.keys())

            error_msg = f"'{self.__class__.__name__}' object has no attribute '{name}'"

            if available_methods:
                error_msg += f". Available methods: {available_methods}"
                raise AttributeError(error_msg) from e
            else:
                error_msg += ". This model is a non-pythonic model. Please use the old inference methods i.e. predict_by_url, predict_by_bytes, etc."
                raise Exception(error_msg) from e

    def _fetch_signatures(self):
        '''
        Fetch the method signatures from the model.

        Returns:
            Dict: The method signatures.
        '''
        try:
            response = self.STUB.GetModelVersion(
                service_pb2.GetModelVersionRequest(
                    user_app_id=self.request_template.user_app_id,
                    model_id=self.request_template.model_id,
                    version_id=self.request_template.version_id,
                )
            )

            method_signatures = None
            if response.status.code == status_code_pb2.SUCCESS:
                method_signatures = response.model_version.method_signatures
            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model failed with response {response!r}")
            self._method_signatures = {}
            for method_signature in method_signatures:
                method_name = method_signature.name
                # check for duplicate method names
                if method_name in self._method_signatures:
                    raise ValueError(f"Duplicate method name {method_name}")
                self._method_signatures[method_name] = method_signature
            if not self._method_signatures:  # if no method signatures, try to fetch from the model
                self._fetch_signatures_backup()
        except Exception:
            # try to fetch from the model
            self._fetch_signatures_backup()
            if not self._method_signatures:
                raise ValueError("Failed to fetch method signatures from model and backup method")

    def _fetch_signatures_backup(self):
        '''
        This is a temporary method of fetching the method signatures from the model.

        Returns:
            Dict: The method signatures.
        '''

        request = service_pb2.PostModelOutputsRequest()
        request.CopyFrom(self.request_template)
        # request.model.model_version.output_info.params['_method_name'] = '_GET_SIGNATURES'
        inp = request.inputs.add()  # empty input for this method
        inp.data.parts.add()  # empty part for this input
        inp.data.metadata['_method_name'] = '_GET_SIGNATURES'
        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        while True:
            response = self.STUB.PostModelOutputs(request)
            if (
                status_is_retryable(response.status.code) and time.time() - start_time < 60 * 10
            ):  # 10 minutes
                logger.info(f"Retrying model info fetch with response {response.status!r}")
                time.sleep(next(backoff_iterator))
                continue
            break
        if response.status.code == status_code_pb2.INPUT_UNSUPPORTED_FORMAT or (
            response.status.code == status_code_pb2.SUCCESS
            and response.outputs[0].data.text.raw == ''
        ):
            # return codes/values from older models that don't support _GET_SIGNATURES
            self._method_signatures = {}
            self._define_compatability_functions()
            return
        if response.status.code != status_code_pb2.SUCCESS:
            if response.outputs[0].status.description.startswith("cannot identify image file"):
                raise Exception(
                    "Failed to fetch method signatures from model and backup method. This model is a non-pythonic model. Please use the old inference methods i.e. predict_by_url, predict_by_bytes, etc."
                )
            raise Exception(f"Model failed with response {response!r}")
        self._method_signatures = signatures_from_json(response.outputs[0].data.text.raw)

    def _define_functions(self):
        '''
        Define the functions based on the method signatures.
        '''
        for method_name, method_signature in self._method_signatures.items():
            # define the function in this client instance
            if resources_pb2.RunnerMethodType.Name(method_signature.method_type) == 'UNARY_UNARY':
                call_func = self._predict
                async_call_func = self._async_predict
            elif (
                resources_pb2.RunnerMethodType.Name(method_signature.method_type)
                == 'UNARY_STREAMING'
            ):
                call_func = self._generate
                async_call_func = self._async_generate
            elif (
                resources_pb2.RunnerMethodType.Name(method_signature.method_type)
                == 'STREAMING_STREAMING'
            ):
                call_func = self._stream
                async_call_func = self._async_stream
            else:
                raise ValueError(f"Unknown method type {method_signature.method_type}")

            # method argnames, in order, collapsing nested keys to corresponding user function args
            method_argnames = []
            for var in method_signature.input_fields:
                outer = var.name.split('.', 1)[0]
                if outer in method_argnames:
                    continue
                method_argnames.append(outer)

            def bind_f(method_name, method_argnames, call_func, async_call_func):
                def sync_f(*args, **kwargs):
                    if len(args) > len(method_argnames):
                        raise TypeError(
                            f"{method_name}() takes {len(method_argnames)} positional arguments but {len(args)} were given"
                        )

                    if len(args) + len(kwargs) > len(method_argnames):
                        raise TypeError(
                            f"{method_name}() got an unexpected keyword argument {next(iter(kwargs))}"
                        )
                    if len(args) == 1 and (not kwargs) and isinstance(args[0], list):
                        batch_inputs = args[0]
                        # Validate each input is a dictionary
                        is_batch_input_valid = all(
                            isinstance(input, dict) for input in batch_inputs
                        )
                        if is_batch_input_valid and (not is_openai_chat_format(batch_inputs)):
                            # If the batch input is valid, call the function with the batch inputs and the method name
                            return call_func(batch_inputs, method_name)

                    for name, arg in zip(
                        method_argnames, args
                    ):  # handle positional with zip shortest
                        if name in kwargs:
                            raise TypeError(f"Multiple values for argument {name}")
                        kwargs[name] = arg
                    return call_func(kwargs, method_name)

                async def async_f(*args, **kwargs):
                    # Async version to call the async function
                    if len(args) > len(method_argnames):
                        raise TypeError(
                            f"{method_name}() takes {len(method_argnames)} positional arguments but {len(args)} were given"
                        )
                    if len(args) + len(kwargs) > len(method_argnames):
                        raise TypeError(
                            f"{method_name}() got an unexpected keyword argument {next(iter(kwargs))}"
                        )
                    if len(args) == 1 and (not kwargs) and isinstance(args[0], list):
                        batch_inputs = args[0]
                        # Validate each input is a dictionary
                        is_batch_input_valid = all(
                            isinstance(input, dict) for input in batch_inputs
                        )
                        if is_batch_input_valid and (not is_openai_chat_format(batch_inputs)):
                            # If the batch input is valid, call the function with the batch inputs and the method name
                            return async_call_func(batch_inputs, method_name)

                    for name, arg in zip(
                        method_argnames, args
                    ):  # handle positional with zip shortest
                        if name in kwargs:
                            raise TypeError(f"Multiple values for argument {name}")
                        kwargs[name] = arg

                    return async_call_func(kwargs, method_name)

                class MethodWrapper:
                    def __call__(self, *args, **kwargs):
                        if is_async_context():
                            return async_f(*args, **kwargs)
                        return sync_f(*args, **kwargs)

                return MethodWrapper()

            # need to bind method_name to the value, not the mutating loop variable
            f = bind_f(method_name, method_argnames, call_func, async_call_func)

            # set names, annotations and docstrings
            f.__name__ = method_name
            f.__qualname__ = f'{self.__class__.__name__}.{method_name}'
            input_annotations = code_script._get_annotations_source(method_signature)
            return_annotation = input_annotations.pop('return', (None, None, None))[0]
            sig = inspect.signature(f).replace(
                parameters=[
                    inspect.Parameter(k, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=v[0])
                    for k, v in input_annotations.items()
                ],
                return_annotation=return_annotation,
            )
            f.__signature__ = sig
            f.__doc__ = method_signature.description
            setattr(self, method_name, f)

    def available_methods(self) -> List[str]:
        """Get the available methods for this model.

        Returns:
            List[str]: The available methods.
        """
        if not self._defined:
            self.fetch()
        return self._method_signatures.keys()

    def method_signature(self, method_name: str) -> str:
        """Get the method signature for a method.

        Args:
            method_name (str): The name of the method.

        Returns:
            str: The method signature.
        """
        if not self._defined:
            self.fetch()
        return method_signatures.get_method_signature(self._method_signatures[method_name])

    def generate_client_script(
        self,
        base_url: str = None,
        use_ctx: bool = False,
        colorize: bool = False,
    ) -> str:
        """Generate a client script for this model.

        Returns:
            str: The client script.
        """
        if not self._defined:
            self.fetch()
        method_signatures = []
        for _, method_signature in self._method_signatures.items():
            method_signatures.append(method_signature)
        return code_script.generate_client_script(
            method_signatures,
            user_id=self.request_template.user_app_id.user_id,
            app_id=self.request_template.user_app_id.app_id,
            model_id=self.request_template.model_id,
            base_url=base_url,
            deployment_id=self.request_template.runner_selector.deployment.id,
            compute_cluster_id=self.request_template.runner_selector.nodepool.compute_cluster.id,
            nodepool_id=self.request_template.runner_selector.nodepool.id,
            use_ctx=use_ctx,
            colorize=colorize,
        )

    def _define_compatability_functions(self):
        serializer = CompatibilitySerializer()

        def predict(input: Any) -> Any:
            proto = resources_pb2.Input()
            serializer.serialize(proto.data, input)
            # always use text.raw for compat
            if proto.data.string_value:
                proto.data.text.raw = proto.data.string_value
                proto.data.string_value = ''
            response = self._predict_by_proto([proto])
            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model predict failed with response {response!r}")
            response_data = response.outputs[0].data
            if response_data.text.raw:
                response_data.string_value = response_data.text.raw
                response_data.text.raw = ''
            return serializer.deserialize(response_data)

        self.predict = predict

    def _predict(
        self,
        inputs,  # TODO set up functions according to fetched signatures?
        method_name: str = 'predict',
    ) -> Any:
        input_signature = self._method_signatures[method_name].input_fields
        output_signature = self._method_signatures[method_name].output_fields

        batch_input = True
        if isinstance(inputs, dict):
            inputs = [inputs]
            batch_input = False

        proto_inputs = []
        for input in inputs:
            proto = resources_pb2.Input()

            serialize(input, input_signature, proto.data)
            proto_inputs.append(proto)

        response = self._predict_by_proto(proto_inputs, method_name)

        outputs = []
        for output in response.outputs:
            outputs.append(deserialize(output.data, output_signature, is_output=True))
        if batch_input:
            return outputs
        return outputs[0]

    def _predict_by_proto(
        self,
        inputs: List[resources_pb2.Input],
        method_name: str = None,
        inference_params: Dict = None,
        output_config: Dict = None,
    ) -> service_pb2.MultiOutputResponse:
        """Predicts the model based on the given inputs.

        Args:
            inputs (List[resources_pb2.Input]): The inputs to predict.
            method_name (str): The remote method name to call.
            inference_params (Dict): Inference parameters to override.
            output_config (Dict): Output configuration to override.

        Returns:
            service_pb2.MultiOutputResponse: The prediction response(s).
        """
        if not isinstance(inputs, list):
            raise UserError('Invalid inputs, inputs must be a list of Input objects.')
        if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
            raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}.")

        request = service_pb2.PostModelOutputsRequest()
        request.CopyFrom(self.request_template)

        request.inputs.extend(inputs)

        if method_name:
            # TODO put in new proto field?
            for inp in request.inputs:
                inp.data.metadata['_method_name'] = method_name
        if inference_params:
            request.model.model_version.output_info.params.update(inference_params)
        if output_config:
            request.model.model_version.output_info.output_config.MergeFrom(
                resources_pb2.OutputConfig(**output_config)
            )

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        while True:
            response = self.STUB.PostModelOutputs(request)
            if (
                status_is_retryable(response.status.code) and time.time() - start_time < 60 * 10
            ):  # 10 minutes
                logger.info("Model is still deploying, please wait...")
                time.sleep(next(backoff_iterator))
                continue

            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model predict failed with response {response!r}")
            break
        return response

    async def _async_predict(
        self,
        inputs,
        method_name: str = 'predict',
    ) -> Any:
        """Asynchronously process inputs and make predictions.

        Args:
            inputs: Input data to process
            method_name (str): Name of the method to call

        Returns:
            Processed prediction results
        """
        # method_name is set to 'predict' by default, this is because to replicate the input and output signature behaviour of sync to async predict.
        input_signature = self._method_signatures[method_name].input_fields
        output_signature = self._method_signatures[method_name].output_fields

        batch_input = True
        if isinstance(inputs, dict):
            inputs = [inputs]
            batch_input = False

        proto_inputs = []
        for input in inputs:
            proto = resources_pb2.Input()
            serialize(input, input_signature, proto.data)
            proto_inputs.append(proto)
        response = await self._async_predict_by_proto(proto_inputs, method_name)
        outputs = []
        for output in response.outputs:
            outputs.append(deserialize(output.data, output_signature, is_output=True))

        return outputs if batch_input else outputs[0]

    async def _async_predict_by_proto(
        self,
        inputs: List[resources_pb2.Input],
        method_name: str = None,
        inference_params: Dict = None,
        output_config: Dict = None,
    ) -> service_pb2.MultiOutputResponse:
        """Asynchronously predicts the model based on the given inputs.

        Args:
            inputs (List[resources_pb2.Input]): The inputs to predict.
            method_name (str): The remote method name to call.
            inference_params (Dict): Inference parameters to override.
            output_config (Dict): Output configuration to override.

        Returns:
            service_pb2.MultiOutputResponse: The prediction response(s).

        Raises:
            UserError: If inputs are invalid or exceed maximum limit.
            Exception: If the model prediction fails.
        """
        if not isinstance(inputs, list):
            raise UserError('Invalid inputs, inputs must be a list of Input objects.')
        if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
            raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}.")

        request = service_pb2.PostModelOutputsRequest()
        request.CopyFrom(self.request_template)
        request.inputs.extend(inputs)

        if method_name:
            for inp in request.inputs:
                inp.data.metadata['_method_name'] = method_name
        if inference_params:
            request.model.model_version.output_info.params.update(inference_params)
        if output_config:
            request.model.model_version.output_info.output_config.MergeFrom(
                resources_pb2.OutputConfig(**output_config)
            )

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)

        while True:
            try:
                response = await self.async_stub.PostModelOutputs(request)

                if (
                    status_is_retryable(response.status.code)
                    and time.time() - start_time < 60 * 10
                ):  # 10 minutes
                    logger.info("Model is still deploying, please wait...")
                    await asyncio.sleep(next(backoff_iterator))
                    continue

                if response.status.code != status_code_pb2.SUCCESS:
                    raise Exception(f"Model predict failed with response {response!r}")

                return response

            except Exception as e:
                if time.time() - start_time >= 10 * 1:  # 10 minutes timeout
                    raise Exception("Model prediction timed out after 10 minutes") from e
                logger.error(f"Error during prediction: {e}")
                await asyncio.sleep(next(backoff_iterator))
                continue

    def _generate(
        self,
        inputs,  # TODO set up functions according to fetched signatures?
        method_name: str = 'generate',
    ) -> Any:
        input_signature = self._method_signatures[method_name].input_fields
        output_signature = self._method_signatures[method_name].output_fields

        batch_input = True
        if isinstance(inputs, dict):
            inputs = [inputs]
            batch_input = False

        proto_inputs = []
        for input in inputs:
            proto = resources_pb2.Input()
            serialize(input, input_signature, proto.data)
            proto_inputs.append(proto)

        response_stream = self._generate_by_proto(proto_inputs, method_name)

        for response in response_stream:
            outputs = []
            for output in response.outputs:
                outputs.append(deserialize(output.data, output_signature, is_output=True))
            if batch_input:
                yield outputs
            else:
                yield outputs[0]

    def _generate_by_proto(
        self,
        inputs: List[resources_pb2.Input],
        method_name: str = None,
        inference_params: Dict = {},
        output_config: Dict = {},
    ):
        """Generate the stream output on model based on the given inputs.

        Args:
            inputs (list[Input]): The inputs to generate, must be less than 128.
            method_name (str): The remote method name to call.
            inference_params (dict): The inference params to override.
            output_config (dict): The output config to override.
        """
        if not isinstance(inputs, list):
            raise UserError('Invalid inputs, inputs must be a list of Input objects.')
        if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
            raise UserError(
                f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}."
            )  # TODO Use Chunker for inputs len > 128

        request = service_pb2.PostModelOutputsRequest()
        request.CopyFrom(self.request_template)

        request.inputs.extend(inputs)

        if method_name:
            # TODO put in new proto field?
            for inp in request.inputs:
                inp.data.metadata['_method_name'] = method_name
        if inference_params:
            request.model.model_version.output_info.params.update(inference_params)
        if output_config:
            request.model.model_version.output_info.output_config.MergeFromDict(output_config)

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        started = False
        while not started:
            stream_response = self.STUB.GenerateModelOutputs(request)
            try:
                response = next(stream_response)  # get the first response
            except StopIteration:
                raise Exception("Model Generate failed with no response")
            if status_is_retryable(response.status.code) and time.time() - start_time < 60 * 10:
                logger.info("Model is still deploying, please wait...")
                time.sleep(next(backoff_iterator))
                continue
            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model Generate failed with response {response.status!r}")
            started = True

        yield response  # yield the first response

        for response in stream_response:
            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model Generate failed with response {response.status!r}")
            yield response

    async def _async_generate(
        self,
        inputs,
        method_name: str = 'generate',
    ) -> Any:
        # method_name is set to 'generate' by default, this is because to replicate the input and output signature behaviour of sync to async generate.
        input_signature = self._method_signatures[method_name].input_fields
        output_signature = self._method_signatures[method_name].output_fields

        batch_input = True
        if isinstance(inputs, dict):
            inputs = [inputs]
            batch_input = False

        proto_inputs = []
        for input in inputs:
            proto = resources_pb2.Input()
        serialize(input, input_signature, proto.data)
        proto_inputs.append(proto)
        response_stream = self._async_generate_by_proto(proto_inputs, method_name)

        async for response in response_stream:
            outputs = []
            for output in response.outputs:
                outputs.append(deserialize(output.data, output_signature, is_output=True))
            if batch_input:
                yield outputs
            else:
                yield outputs[0]

    async def _async_generate_by_proto(
        self,
        inputs: List[resources_pb2.Input],
        method_name: str = None,
        inference_params: Dict = {},
        output_config: Dict = {},
    ):
        """Generate the async stream output on model based on the given inputs.

        Args:
            inputs (list[Input]): The inputs to generate, must be less than 128.
            method_name (str): The remote method name to call.
            inference_params (dict): The inference params to override.
            output_config (dict): The output config to override.
        """
        if not isinstance(inputs, list):
            raise UserError('Invalid inputs, inputs must be a list of Input objects.')
        if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
            raise UserError(
                f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}."
            )  # TODO Use Chunker for inputs len > 128

        request = service_pb2.PostModelOutputsRequest()
        request.CopyFrom(self.request_template)

        request.inputs.extend(inputs)

        if method_name:
            # TODO put in new proto field?
            for inp in request.inputs:
                inp.data.metadata['_method_name'] = method_name
        if inference_params:
            request.model.model_version.output_info.params.update(inference_params)
        if output_config:
            request.model.model_version.output_info.output_config.MergeFromDict(output_config)

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        started = False
        while not started:
            # stream response returns gRPC async iterable - UnaryStreamCall
            stream_response = self.async_stub.GenerateModelOutputs(request)
            stream_resp = await stream_response  # get the async iterable
            iterator = stream_resp.__aiter__()  # get the async iterator for the response
            try:
                response = await iterator.__anext__()  # getting the first response
            except StopAsyncIteration:
                raise Exception("Model Generate failed with no response")
            if status_is_retryable(response.status.code) and time.time() - start_time < 60 * 10:
                logger.info("Model is still deploying, please wait...")
                await asyncio.sleep(next(backoff_iterator))
                continue
            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model Generate failed with response {response.status!r}")
            started = True

        yield response  # yield the first response

        async for response in iterator:
            if response.status.code != status_code_pb2.SUCCESS:
                raise Exception(f"Model Generate failed with response {response.status!r}")
            yield response

    def _stream(
        self,
        inputs,
        method_name: str = 'stream',
    ) -> Any:
        input_signature = self._method_signatures[method_name].input_fields
        output_signature = self._method_signatures[method_name].output_fields

        if isinstance(inputs, list):
            assert len(inputs) == 1, 'streaming methods do not support batched calls'
            inputs = inputs[0]
        assert isinstance(inputs, dict)
        kwargs = inputs

        # find the streaming vars in the input signature, and the streaming input python param
        stream_sig = get_stream_from_signature(input_signature)
        if stream_sig is None:
            raise ValueError("Streaming method must have a Stream input")
        stream_argname = stream_sig.name

        # get the streaming input generator from the user-provided function arg values
        user_inputs_generator = kwargs.pop(stream_argname)

        def _input_proto_stream():
            # first item contains all the inputs and the first stream item
            proto = resources_pb2.Input()
            try:
                item = next(user_inputs_generator)
            except StopIteration:
                return  # no items to stream
            kwargs[stream_argname] = item
            serialize(kwargs, input_signature, proto.data)

            yield proto

            # subsequent items are just the stream items
            for item in user_inputs_generator:
                proto = resources_pb2.Input()
                serialize({stream_argname: item}, [stream_sig], proto.data)
                yield proto

        response_stream = self._stream_by_proto(_input_proto_stream(), method_name)

        for response in response_stream:
            assert len(response.outputs) == 1, 'streaming methods must have exactly one output'
            yield deserialize(response.outputs[0].data, output_signature, is_output=True)

    def _req_iterator(
        self,
        input_iterator: Iterator[List[resources_pb2.Input]],
        method_name: str = None,
        inference_params: Dict = {},
        output_config: Dict = {},
    ):
        request = service_pb2.PostModelOutputsRequest()
        request.CopyFrom(self.request_template)
        if inference_params:
            request.model.model_version.output_info.params.update(inference_params)
        if output_config:
            request.model.model_version.output_info.output_config.MergeFromDict(output_config)
        for inputs in input_iterator:
            req = service_pb2.PostModelOutputsRequest()
            req.CopyFrom(request)
            if isinstance(inputs, list):
                req.inputs.extend(inputs)
            else:
                req.inputs.append(inputs)
            # TODO: put into new proto field?
            if method_name:
                for inp in req.inputs:
                    inp.data.metadata['_method_name'] = method_name
            yield req

    def _stream_by_proto(
        self,
        inputs: Iterator[List[resources_pb2.Input]],
        method_name: str = None,
        inference_params: Dict = {},
        output_config: Dict = {},
    ):
        """Generate the stream output on model based on the given stream of inputs."""
        # if not isinstance(inputs, Iterator[List[Input]]):
        #   raise UserError('Invalid inputs, inputs must be a iterator of list of Input objects.')

        request = self._req_iterator(inputs, method_name, inference_params, output_config)

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        generation_started = False
        while True:
            if generation_started:
                break
            stream_response = self.STUB.StreamModelOutputs(request)
            for response in stream_response:
                if (
                    status_is_retryable(response.status.code)
                    and time.time() - start_time < 60 * 10
                ):
                    logger.info("Model is still deploying, please wait...")
                    time.sleep(next(backoff_iterator))
                    break
                if response.status.code != status_code_pb2.SUCCESS:
                    raise Exception(f"Model Predict failed with response {response.status!r}")
                else:
                    if not generation_started:
                        generation_started = True
                    yield response

    # TODO: Test async streaming.
    async def _async_stream(
        self,
        inputs,
        method_name: str = 'stream',
    ) -> Any:
        # method_name is set to 'stream' by default, this is because to replicate the input and output signature behaviour of sync to async stream.
        input_signature = self._method_signatures[method_name].input_fields
        output_signature = self._method_signatures[method_name].output_fields

        if isinstance(inputs, list):
            assert len(inputs) == 1, 'streaming methods do not support batched calls'
            inputs = inputs[0]
        assert isinstance(inputs, dict)
        kwargs = inputs

        # find the streaming vars in the input signature, and the streaming input python param
        stream_sig = get_stream_from_signature(input_signature)
        if stream_sig is None:
            raise ValueError("Streaming method must have a Stream input")
        stream_argname = stream_sig.name

        # get the streaming input generator from the user-provided function arg values
        user_inputs_generator = kwargs.pop(stream_argname)

        async def _input_proto_stream():
            # first item contains all the inputs and the first stream item
            proto = resources_pb2.Input()
            try:
                item = await user_inputs_generator.__anext__()
            except StopAsyncIteration:
                return  # no items to stream
            kwargs[stream_argname] = item
            serialize(kwargs, input_signature, proto.data)

            yield proto

        # subsequent items are just the stream items
        async for item in user_inputs_generator:
            proto = resources_pb2.Input()
            serialize({stream_argname: item}, [stream_sig], proto.data)
            yield proto

        response_stream = await self._async_stream_by_proto(_input_proto_stream(), method_name)

        async for response in response_stream:
            assert len(response.outputs) == 1, 'streaming methods must have exactly one output'
            yield deserialize(response.outputs[0].data, output_signature, is_output=True)

    async def _async_stream_by_proto(
        self,
        inputs: Iterator[List[resources_pb2.Input]],
        method_name: str = None,
        inference_params: Dict = {},
        output_config: Dict = {},
    ):
        """Generate the async stream output on model based on the given stream of inputs."""
        # if not isinstance(inputs, Iterator[List[Input]]):
        #   raise UserError('Invalid inputs, inputs must be a iterator of list of Input objects.')

        request = self._req_iterator(inputs, method_name, inference_params, output_config)

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        generation_started = False
        while True:
            if generation_started:
                break
            stream_response = await self.async_stub.StreamModelOutputs(request)
            async for response in stream_response:
                if (
                    status_is_retryable(response.status.code)
                    and time.time() - start_time < 60 * 10
                ):
                    logger.info("Model is still deploying, please wait...")
                    await asyncio.sleep(next(backoff_iterator))
                    break
                if response.status.code != status_code_pb2.SUCCESS:
                    raise Exception(f"Model Predict failed with response {response.status!r}")
                else:
                    if not generation_started:
                        generation_started = True
                    yield response



import datetime
import json
import logging

from clarifai.utils.logging import (
    JsonFormatter,
    TerminalFormatter,
    _get_library_name,
    get_logger,
    set_logger_context,
)


def test_get_logger():
    logger = get_logger("DEBUG", "test_logger")
    assert logger.level == logging.DEBUG
    assert logger.name == "test_logger"
    assert isinstance(logger.handlers[0], logging.StreamHandler)
    assert isinstance(logger.handlers[0].formatter, TerminalFormatter)


def test_get_logger_defaults():
    logger = get_logger()
    assert logger.level == logging.NOTSET
    assert logger.name == _get_library_name()
    assert isinstance(logger.handlers[0], logging.StreamHandler)
    assert isinstance(logger.handlers[0].formatter, TerminalFormatter)


def test_get_json_logger_defaults(monkeypatch):
    # with env setting of ENABLE_JSON_LOGGER to true
    # we should get a json logger.
    monkeypatch.setenv("ENABLE_JSON_LOGGER", "true")
    logger = get_logger()
    assert logger.level == logging.NOTSET
    assert logger.name == _get_library_name()
    assert isinstance(logger.handlers[0], logging.StreamHandler)
    assert isinstance(logger.handlers[0].formatter, JsonFormatter)


def test_json_logger():
    filename = "testy.py"
    msg = "testy"
    lineno = 1
    sinfo = "testf2"
    r = logging.LogRecord(
        name="testy",
        level=logging.INFO,
        pathname=filename,
        lineno=lineno,
        msg=msg,
        args=(),
        exc_info=None,
        func="testf",
        sinfo=sinfo,
    )
    jf = JsonFormatter()
    # format the record as a json line.
    json_line = jf.format(r)
    result = json.loads(json_line)
    # parse timestamp of format "2024-09-24T22:06:49.573038Z" in @timestamp field.
    assert result["@timestamp"] is not None
    ts = result["@timestamp"]
    # assert the ts was within 10 seconds of now (in UTC time)
    assert abs(
        datetime.datetime.utcnow() - datetime.datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%fZ")
    ) < datetime.timedelta(seconds=10)
    assert result['filename'] == filename
    assert result['msg'] == msg
    assert result['lineno'] == lineno
    assert result['stack_info'] == sinfo
    assert result['level'] == "info"
    assert 'req_id' not in result

    req_id = "test_req_id"
    set_logger_context(req_id=req_id)
    json_line = jf.format(r)
    result = json.loads(json_line)
    assert abs(
        datetime.datetime.utcnow() - datetime.datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%fZ")
    ) < datetime.timedelta(seconds=10)
    assert result['filename'] == filename
    assert result['msg'] == msg
    assert result['lineno'] == lineno
    assert result['stack_info'] == sinfo
    assert result['level'] == "info"
    assert result['req_id'] == req_id



import inspect
import itertools
import os
import threading
import traceback
from abc import ABC
from collections import abc
from typing import Any, Dict, Iterator, List
from unittest.mock import MagicMock

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2

from clarifai.runners.utils import data_types
from clarifai.runners.utils.data_utils import DataConverter
from clarifai.runners.utils.method_signatures import (
    build_function_signature,
    deserialize,
    get_stream_from_signature,
    serialize,
    signatures_to_json,
)
from clarifai.runners.utils.model_utils import is_proto_style_method
from clarifai.utils.logging import logger

_METHOD_INFO_ATTR = '_cf_method_info'

_RAISE_EXCEPTIONS = os.getenv("RAISE_EXCEPTIONS", "false").lower() in ("true", "1")

FALLBACK_METHOD_PROTO = 'PostModelOutputs'
FALLBACK_METHOD_PYTHON = 'predict'


class ModelClass(ABC):
    '''
    Base class for model classes that can be run as a service.

    Define predict, generate, or stream methods using the @ModelClass.method decorator.

    Example:

      from clarifai.runners.model_class import ModelClass
      from clarifai.runners.utils.data_types import NamedFields
      from typing import List, Iterator

      class MyModel(ModelClass):

        @ModelClass.method
        def predict(self, x: str, y: int) -> List[str]:
          return [x] * y

        @ModelClass.method
        def generate(self, x: str, y: int) -> Iterator[str]:
          for i in range(y):
            yield x + str(i)

        @ModelClass.method
        def stream(self, input_stream: Iterator[NamedFields(x=str, y=int)]) -> Iterator[str]:
          for item in input_stream:
            yield item.x + ' ' + str(item.y)
    '''

    def __init__(self):
        super().__init__()
        self._thread_local = threading.local()

    @staticmethod
    def method(func):
        setattr(func, _METHOD_INFO_ATTR, _MethodInfo(func))
        return func

    def set_output_context(self, prompt_tokens=None, completion_tokens=None):
        """Set the prompt and completion tokens for the Output proto.
        In batch mode, call this once per output, in order, before returning each output.
        """
        if not hasattr(self._thread_local, 'token_contexts'):
            self._thread_local.token_contexts = []
        self._thread_local.token_contexts.append((prompt_tokens, completion_tokens))

    def load_model(self):
        """Load the model."""

    def _handle_get_signatures_request(self) -> service_pb2.MultiOutputResponse:
        methods = self._get_method_infos()
        signatures = {
            method.name: method.signature
            for method in methods.values()
            if method.signature is not None
        }
        resp = service_pb2.MultiOutputResponse(
            status=status_pb2.Status(code=status_code_pb2.SUCCESS)
        )
        output = resp.outputs.add()
        output.status.code = status_code_pb2.SUCCESS
        output.data.text.raw = signatures_to_json(signatures)
        return resp

    def _batch_predict(self, method, inputs: List[Dict[str, Any]]) -> List[Any]:
        """Batch predict method for multiple inputs."""
        outputs = []
        for input in inputs:
            output = method(**input)
            outputs.append(output)
        return outputs

    def _batch_generate(self, method, inputs: List[Dict[str, Any]]) -> Iterator[List[Any]]:
        """Batch generate method for multiple inputs."""
        generators = [method(**input) for input in inputs]
        for outputs in itertools.zip_longest(*generators):
            yield outputs

    def predict_wrapper(
        self, request: service_pb2.PostModelOutputsRequest
    ) -> service_pb2.MultiOutputResponse:
        outputs = []
        try:
            # TODO add method name field to proto
            # to support old callers who might not pass in the method name we have a few defaults.
            # first we look for a PostModelOutputs method that is implemented as protos and use that
            # if it exists.
            # if not we default to 'predict'.
            method_name = None
            if len(request.inputs) > 0 and '_method_name' in request.inputs[0].data.metadata:
                method_name = request.inputs[0].data.metadata['_method_name']
            if method_name is None and FALLBACK_METHOD_PROTO in self._get_method_infos():
                _info = self._get_method_infos(FALLBACK_METHOD_PROTO)
                if _info.proto_method:
                    method_name = FALLBACK_METHOD_PROTO
            if method_name is None:
                method_name = FALLBACK_METHOD_PYTHON
            if (
                method_name == '_GET_SIGNATURES'
            ):  # special case to fetch signatures, TODO add endpoint for this
                return self._handle_get_signatures_request()
            if method_name not in self._get_method_infos():
                raise ValueError(f"Method {method_name} not found in model class")
            method = getattr(self, method_name)
            method_info = self._get_method_infos(method_name)
            signature = method_info.signature
            proto_method = method_info.proto_method

            # If this is an old predict(proto) -> proto method, just call it and return
            # the response.
            if proto_method:
                out_proto = method(request)
                # if we already have out_proto.status.code set then return
                if out_proto.status.code != status_code_pb2.ZERO:
                    return out_proto

                successes = [
                    out.status.code == status_code_pb2.SUCCESS for out in out_proto.outputs
                ]
                if all(successes):
                    # If all outputs are successful, we can return the response.
                    out_proto.status.CopyFrom(
                        status_pb2.Status(code=status_code_pb2.SUCCESS, description='Success')
                    )
                    return out_proto
                if any(successes):
                    # If some outputs are successful and some are not, we return a mixed status.
                    out_proto.status.CopyFrom(
                        status_pb2.Status(
                            code=status_code_pb2.MIXED_STATUS, description='Mixed Status'
                        )
                    )
                    return out_proto
                # If all outputs are failures, we return a failure status.
                out_proto.status.CopyFrom(
                    status_pb2.Status(code=status_code_pb2.FAILURE, description='Failed')
                )
                return out_proto

            python_param_types = method_info.python_param_types
            for input in request.inputs:
                # check if input is in old format
                is_convert = DataConverter.is_old_format(input.data)
                if is_convert:
                    # convert to new format
                    new_data = DataConverter.convert_input_data_to_new_format(
                        input.data, signature.input_fields
                    )
                    input.data.CopyFrom(new_data)

            # convert inputs to python types
            inputs = self._convert_input_protos_to_python(
                request.inputs, signature.input_fields, python_param_types
            )
            if len(inputs) == 1:
                inputs = inputs[0]
                output = method(**inputs)
                outputs.append(
                    self._convert_output_to_proto(
                        output, signature.output_fields, convert_old_format=is_convert
                    )
                )
            else:
                outputs = self._batch_predict(method, inputs)
                outputs = [
                    self._convert_output_to_proto(
                        output, signature.output_fields, convert_old_format=is_convert
                    )
                    for output in outputs
                ]

            return service_pb2.MultiOutputResponse(
                outputs=outputs, status=status_pb2.Status(code=status_code_pb2.SUCCESS)
            )
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            logger.exception("Error in predict")
            return service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.FAILURE,
                    details=str(e),
                    stack_trace=traceback.format_exc().split('\n'),
                )
            )

    def generate_wrapper(
        self, request: service_pb2.PostModelOutputsRequest
    ) -> Iterator[service_pb2.MultiOutputResponse]:
        try:
            assert len(request.inputs) == 1, "Generate requires exactly one input"
            method_name = 'generate'
            if len(request.inputs) > 0 and '_method_name' in request.inputs[0].data.metadata:
                method_name = request.inputs[0].data.metadata['_method_name']
            method = getattr(self, method_name)
            method_info = self._get_method_infos(method_name)
            signature = method_info.signature
            python_param_types = method_info.python_param_types
            for input in request.inputs:
                # check if input is in old format
                is_convert = DataConverter.is_old_format(input.data)
                if is_convert:
                    # convert to new format
                    new_data = DataConverter.convert_input_data_to_new_format(
                        input.data, signature.input_fields
                    )
                    input.data.CopyFrom(new_data)
            inputs = self._convert_input_protos_to_python(
                request.inputs, signature.input_fields, python_param_types
            )
            if len(inputs) == 1:
                inputs = inputs[0]
                for output in method(**inputs):
                    resp = service_pb2.MultiOutputResponse()
                    self._convert_output_to_proto(
                        output,
                        signature.output_fields,
                        proto=resp.outputs.add(),
                        convert_old_format=is_convert,
                    )
                    resp.status.code = status_code_pb2.SUCCESS
                    yield resp
            else:
                for outputs in self._batch_generate(method, inputs):
                    resp = service_pb2.MultiOutputResponse()
                    for output in outputs:
                        self._convert_output_to_proto(
                            output,
                            signature.output_fields,
                            proto=resp.outputs.add(),
                            convert_old_format=is_convert,
                        )
                    resp.status.code = status_code_pb2.SUCCESS
                    yield resp
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            logger.exception("Error in generate")
            yield service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.FAILURE,
                    details=str(e),
                    stack_trace=traceback.format_exc().split('\n'),
                )
            )

    def stream_wrapper(
        self, request_iterator: Iterator[service_pb2.PostModelOutputsRequest]
    ) -> Iterator[service_pb2.MultiOutputResponse]:
        try:
            request = next(request_iterator)  # get first request to determine method
            assert len(request.inputs) == 1, "Streaming requires exactly one input"

            method_name = 'stream'
            if len(request.inputs) > 0 and '_method_name' in request.inputs[0].data.metadata:
                method_name = request.inputs[0].data.metadata['_method_name']
            method = getattr(self, method_name)
            method_info = self._get_method_infos(method_name)
            signature = method_info.signature
            python_param_types = method_info.python_param_types

            # find the streaming vars in the signature
            stream_sig = get_stream_from_signature(signature.input_fields)
            if stream_sig is None:
                raise ValueError("Streaming method must have a Stream input")
            stream_argname = stream_sig.name

            for input in request.inputs:
                # check if input is in old format
                is_convert = DataConverter.is_old_format(input.data)
                if is_convert:
                    # convert to new format
                    new_data = DataConverter.convert_input_data_to_new_format(
                        input.data, signature.input_fields
                    )
                    input.data.CopyFrom(new_data)
            # convert all inputs for the first request, including the first stream value
            inputs = self._convert_input_protos_to_python(
                request.inputs, signature.input_fields, python_param_types
            )
            kwargs = inputs[0]

            # first streaming item
            first_item = kwargs.pop(stream_argname)

            # streaming generator
            def InputStream():
                yield first_item
                # subsequent streaming items contain only the streaming input
                for request in request_iterator:
                    item = self._convert_input_protos_to_python(
                        request.inputs, [stream_sig], python_param_types
                    )
                    item = item[0][stream_argname]
                    yield item

            # add stream generator back to the input kwargs
            kwargs[stream_argname] = InputStream()

            for output in method(**kwargs):
                resp = service_pb2.MultiOutputResponse()
                self._convert_output_to_proto(
                    output,
                    signature.output_fields,
                    proto=resp.outputs.add(),
                    convert_old_format=is_convert,
                )
                resp.status.code = status_code_pb2.SUCCESS
                yield resp
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            logger.exception("Error in stream")
            yield service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.FAILURE,
                    details=str(e),
                    stack_trace=traceback.format_exc().split('\n'),
                )
            )

    def _convert_input_protos_to_python(
        self,
        inputs: List[resources_pb2.Input],
        variables_signature: List[resources_pb2.ModelTypeField],
        python_param_types,
    ) -> List[Dict[str, Any]]:
        result = []
        for input in inputs:
            kwargs = deserialize(input.data, variables_signature)
            # dynamic cast to annotated types
            for k, v in kwargs.items():
                if k not in python_param_types:
                    continue

                if hasattr(python_param_types[k], "__args__") and (
                    getattr(python_param_types[k], "__origin__", None)
                    in [abc.Iterator, abc.Generator, abc.Iterable]
                ):
                    # get the type of the items in the stream
                    stream_type = python_param_types[k].__args__[0]

                    kwargs[k] = data_types.cast(v, stream_type)
                else:
                    kwargs[k] = data_types.cast(v, python_param_types[k])
            result.append(kwargs)
        return result

    def _convert_output_to_proto(
        self,
        output: Any,
        variables_signature: List[resources_pb2.ModelTypeField],
        proto=None,
        convert_old_format=False,
    ) -> resources_pb2.Output:
        if proto is None:
            proto = resources_pb2.Output()
        serialize({'return': output}, variables_signature, proto.data, is_output=True)
        if convert_old_format:
            # convert to old format
            data = DataConverter.convert_output_data_to_old_format(proto.data)
            proto.data.CopyFrom(data)
        proto.status.code = status_code_pb2.SUCCESS
        # Per-output token context support
        token_contexts = getattr(self._thread_local, 'token_contexts', None)
        prompt_tokens = completion_tokens = None
        if token_contexts and len(token_contexts) > 0:
            prompt_tokens, completion_tokens = token_contexts.pop(0)
            # If this was the last, clean up
            if len(token_contexts) == 0:
                del self._thread_local.token_contexts
        if prompt_tokens is not None:
            proto.prompt_tokens = prompt_tokens
        if completion_tokens is not None:
            proto.completion_tokens = completion_tokens
        return proto

    @classmethod
    def _register_model_methods(cls):
        # go up the class hierarchy to find all decorated methods, and add to registry of current class
        methods = {}
        for base in reversed(cls.__mro__):
            for name, method in base.__dict__.items():
                # Skip class attributes, mocked objects, and non-methods
                if not callable(method) or isinstance(method, (classmethod, staticmethod)):
                    continue
                # Skip any mocked objects or attributes
                if isinstance(method, MagicMock) or hasattr(method, '_mock_return_value'):
                    continue
                # Only include methods that have been decorated with @ModelClass.method
                method_info = getattr(method, _METHOD_INFO_ATTR, None)
                if not method_info:  # regular function, not a model method
                    continue
                methods[name] = method_info
        # check for generic predict(request) -> response, etc. methods
        # older models never had generate or stream so don't bother with them.
        for name in [FALLBACK_METHOD_PROTO]:  # , 'GenerateModelOutputs', 'StreamModelOutputs'):
            if hasattr(cls, name) and name not in methods:
                method = getattr(cls, name)
                if not callable(method):
                    continue
                if is_proto_style_method(method):
                    # If this is a proto-style method, we can add it to the registry as a special case.
                    methods[name] = _MethodInfo(method, proto_method=True)
        # set method table for this class in the registry
        return methods

    @classmethod
    def _get_method_infos(cls, func_name=None):
        # FIXME: this is a re-use of the _METHOD_INFO_ATTR attribute to store the method info
        # for all methods on the class. Should use a different attribute name to avoid confusion.
        if not hasattr(cls, _METHOD_INFO_ATTR):
            setattr(cls, _METHOD_INFO_ATTR, cls._register_model_methods())
        method_infos = getattr(cls, _METHOD_INFO_ATTR)
        if func_name:
            return method_infos[func_name]
        return method_infos


class _MethodInfo:
    def __init__(self, method, proto_method=False):
        """Initialize a MethodInfo instance.

        Args:
            method: The method to wrap.
            old_method: If True, this is an old proto-style method that returns a proto directly.
        """
        self.name = method.__name__
        self.proto_method = proto_method
        if not proto_method:
            self.signature = build_function_signature(method)
        else:
            self.signature = None
        self.python_param_types = {
            p.name: p.annotation
            for p in inspect.signature(method).parameters.values()
            if p.annotation != inspect.Parameter.empty
        }
        self.python_param_types.pop('self', None)



import logging
import os
import time
import uuid

import pytest
import yaml

from clarifai.client.user import User

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
NOW = uuid.uuid4().hex[:10]
CREATE_APP_ID = f"ci_input_app_{NOW}"
CREATE_DATASET_ID = "ci_input_test_dataset"
CREATE_MODEL_ID = "ci_input_test_model"
CREATE_MODEL_ID_1 = "ci_input_test_model_1"
CSV_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.csv"

CLARIFAI_API_BASE = os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com")


def client():
    return User(user_id=CREATE_APP_USER_ID, base_url=CLARIFAI_API_BASE)


@pytest.mark.requires_secrets
class Testmodeltrain:
    """Tests for model training."""

    @classmethod
    def setup_class(self):
        self.app = client().create_app(app_id=CREATE_APP_ID, base_workflow="Empty")
        self.input_object = self.app.inputs()
        self.dataset = self.app.create_dataset(dataset_id=CREATE_DATASET_ID)
        self.visual_classifier_model = self.app.create_model(
            model_id=CREATE_MODEL_ID, model_type_id='visual-classifier'
        )
        self.text_classifier_model = self.app.create_model(
            model_id=CREATE_MODEL_ID_1, model_type_id='text-classifier'
        )

    def test_model_templates(self):
        model_types = self.app.list_trainable_model_types()
        templates = self.visual_classifier_model.list_training_templates()
        assert (
            self.visual_classifier_model.model_type_id == 'visual-classifier'
        )  # create model test
        assert len(model_types) == 8  # list trainable model types test
        assert len(templates) >= 11  # list training templates test

    def test_model_params(self):
        model_params = self.visual_classifier_model.get_params(
            template='MMClassification_AdvancedConfig', save_to='tests/assets/model_params.yaml'
        )
        with open('tests/assets/model_params.yaml', 'r') as file:
            expected_data = yaml.safe_load(file)
        with open(expected_data['train_params']['custom_config'], 'r') as python_file:
            custom_config = python_file.read()

        assert (
            expected_data['train_params'].keys() == model_params['train_params'].keys()
        )  # test yaml model params
        assert len(model_params['concepts']) == 0  # test model params concepts
        assert (
            'custom_config' in expected_data['train_params'].keys()
        )  # test custom config in yaml
        assert (
            expected_data['train_params']['template'] == 'MMClassification_AdvancedConfig'
        )  # test template param in yaml
        assert (
            'image_size' in expected_data['train_params'].keys()
        )  # test image size param in yaml
        assert 'dataset_id' in expected_data.keys()  # test dataset id param in yaml
        assert '_base_' in custom_config  # test custom config script
        assert 'max_epochs' in custom_config  # test custom config script
        # cleanup
        os.remove('tests/assets/model_params.yaml')
        os.remove(expected_data['train_params']['custom_config'])

    def test_model_train(self, caplog):
        self.dataset.upload_from_csv(
            csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True
        )
        dataset_version = self.dataset.create_version()
        concepts = [concept.id for concept in self.app.list_concepts()]
        self.text_classifier_model.get_params(
            template='HF_GPTNeo_125m_lora', save_to='tests/assets/model_params.yaml'
        )
        param_info = self.text_classifier_model.get_param_info(param='tokenizer_config')
        assert param_info['param'] == 'tokenizer_config'  # test get param info
        assert len(concepts) == 2  # test data upload for training
        self.text_classifier_model.update_params(
            dataset_id=CREATE_DATASET_ID,
            concepts=concepts,
            dataset_version_id=dataset_version.version.id,
        )
        with caplog.at_level(logging.INFO):
            model_version_id = self.text_classifier_model.train()
            assert "Model Training Started" in caplog.text  # test model training

        training_status = self.text_classifier_model.training_status(version_id=model_version_id)
        assert model_version_id  # test model version id from model.train()
        assert training_status.code  # test training_status
        # NOTE (EAGLE-4139) - Immediately deleting the app causes the training to fail because the first
        # step of training requires an app, this triggers infra alerts for non graceful exit. We need to delay the deletion
        # until we get past this step.

        time.sleep(5)

        with caplog.at_level(logging.INFO):
            self.text_classifier_model.delete_version(version_id=model_version_id)
            assert "Model Version Deleted" in caplog.text  # test model version deletion
        # cleanup
        os.remove('tests/assets/model_params.yaml')

    @classmethod
    def teardown_class(self):
        self.app.delete_dataset(dataset_id=CREATE_DATASET_ID)
        self.app.delete_model(model_id=CREATE_MODEL_ID)
        self.app.delete_model(model_id=CREATE_MODEL_ID_1)
        client().delete_app(app_id=CREATE_APP_ID)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface to Clarifai Models API."""

from typing import Dict, List, Type

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct, Value

from clarifai.client import create_stub
from clarifai.client.auth.helper import ClarifaiAuthHelper


def _make_default_value_proto(dtype, value):
    if dtype == 1:
        return Value(bool_value=value)
    elif dtype == 2 or dtype == 21:
        return Value(string_value=value)
    elif dtype == 3:
        return Value(number_value=value)


class Models:
    """
    Interface to Clarifai models api
    """

    def __init__(self, auth: Type[ClarifaiAuthHelper]) -> None:
        self.auth = auth
        self.stub = create_stub(self.auth)

    def list_model_types(self) -> Dict:
        """
        List all API Model Types that support input and output.

        Returns:
          - A dict of;
              model_type, [{expected_input_name: value}, {expected_output_name: value}, model_descripton]
            key, value pairs respectively.i.e.
              {model_type: [{inp_field_name: value}, {output_field_name: value,...}, model_description]}

          - model_type: supported input shapes and data types dict. Structure;
              {model_type: [(supported_input_N_dims, dtypes_N),...]}

          - model_type: supported output shapes and data types dict. Structure;
              {model_type: [(supported_output_N_dims, dtypes_N),...]}
        """
        ## List model types from API
        model_types_response = self.stub.ListModelTypes(
            service_pb2.ListModelTypesRequest(),  # (user_app_id=auth.user_app_id),
            metadata=self.auth.metadata,
        )
        # model types dict structure:
        # {model_type: [{inp_field_name: value,}, {output_field_name: value,...}, desc]}
        model_types = {}
        in_dims_dtype = {}  # {model_type: [(supported_input_N_dims, dtypes_N),...]}
        out_dims_dtype = {}  # {model_type: [(supported_output_N_dims, dtypes_N),...]}
        types_dict = MessageToDict(model_types_response, preserving_proto_field_name=True)[
            'model_types'
        ]
        for i in range(len(types_dict)):
            model_id = types_dict[i]['id']
            model_desc = types_dict[i]['description']
            if 'expected_output_layers' in types_dict[i].keys():
                # expected_input_layers exist for all expected_output_layers
                # hence one conditional check
                model_types[model_id] = []
                in_dims_dtype[model_id] = []
                out_dims_dtype[model_id] = []
                expected_input = types_dict[i]['expected_input_layers']
                expected_output = types_dict[i]['expected_output_layers']
                for inp in expected_input:
                    if 'data_field_name' in inp.keys():
                        model_types[model_id].append({inp['data_field_name']: None})
                    if 'shapes' in inp.keys():
                        for dim in inp['shapes']:
                            if 'dims' in dim.keys():
                                if 'max_dims' in dim.keys():
                                    in_dims_dtype[model_id].append(
                                        (dim['dims'], dim['max_dims'], dim['data_type'])
                                    )
                                else:
                                    in_dims_dtype[model_id].append((dim['dims'], dim['data_type']))
                    continue
                temp_out = {}
                for each in expected_output:
                    if 'data_field_name' in each.keys():
                        temp_out[each['data_field_name']] = None
                    if 'shapes' in each.keys():
                        for dim in each['shapes']:
                            if 'dims' in dim.keys():
                                out_dims_dtype[model_id].append((dim['dims'], dim['data_type']))
                    else:
                        continue
                model_types[model_id].append(temp_out)
                model_types[model_id].append(model_desc)

        return {
            "Model Types": model_types,
            "Input Metadata": in_dims_dtype,
            "Output Metadata": out_dims_dtype,
        }

    def init_model(
        self,
        model_id: str,
        model_type: str,
        description: str = "",
    ):
        """Init a new model on Clarifai platform.

        Args:
            model_id (str): Clarifai model id
            model_type (str): Clarifai model type
            description (str, optional): a description of the model. Defaults to "".

        Returns:
            dict: Clarifai api response
        """
        user_data_object = self.auth.get_user_app_id_proto()
        post_models_response = self.stub.PostModels(
            service_pb2.PostModelsRequest(
                user_app_id=user_data_object,
                models=[
                    resources_pb2.Model(id=model_id, notes=description, model_type_id=model_type)
                ],
            ),
            metadata=self.auth.metadata,
        )

        return MessageToDict(post_models_response, preserving_proto_field_name=True)

    def post_model_version(
        self,
        model_id: str,
        model_zip_url: str,
        input: dict,
        outputs: dict,
        param_specs: List[dict] = None,
    ):
        """Post a new version of an existing model in the Clarifai platform.

        Args:
            model_id (str): Clarifai model id
            model_zip_url (str]): url of zip of model
            model_zip_url (str): url of zip of model
            input (dict): a dict where the key is clarifai input field and the value is triton model input,
                {clarifai_input_field: triton_input_filed}.
            outputs (dict): a dict where the keys are clarifai output fields and the values are triton model outputs,
                {clarifai_output_field1: triton_output_filed1, clarifai_output_field2: triton_output_filed2,...}.
            param_specs (List[dict]): list of dicts - keys are path, field_type, default_value, description. Default is None
        Returns:
            dict: clarifai api response
        """
        user_data_object = self.auth.get_user_app_id_proto()

        def _parse_fields_map(x):
            """parse input, outputs to Struct"""
            _fields_map = Struct()
            _fields_map.update(x)
            return _fields_map

        input_fields_map = _parse_fields_map(input)
        output_fields_map = _parse_fields_map(outputs)
        # resources_pb2.ModelTypeField(path="abc", default_value=1, description="test")
        if param_specs:
            iterative_proto_params = []
            for param in param_specs:
                dtype = param.get("field_type")
                proto_param = resources_pb2.ModelTypeField(
                    path=param.get("path"),
                    field_type=dtype,
                    default_value=_make_default_value_proto(
                        dtype=dtype, value=param.get("default_value")
                    ),
                    description=param.get("description"),
                )
                iterative_proto_params.append(proto_param)
            param_specs = iterative_proto_params

        post_model_versions = self.stub.PostModelVersions(
            service_pb2.PostModelVersionsRequest(
                user_app_id=user_data_object,
                model_id=model_id,
                model_versions=[
                    resources_pb2.ModelVersion(
                        pretrained_model_config=resources_pb2.PretrainedModelConfig(
                            model_zip_url=model_zip_url,
                            input_fields_map=input_fields_map,
                            output_fields_map=output_fields_map,
                        ),
                        output_info=resources_pb2.OutputInfo(params_specs=param_specs),
                    )
                ],
            ),
            metadata=self.auth.metadata,
        )

        return MessageToDict(post_model_versions, preserving_proto_field_name=True)

    def upload_model(
        self,
        model_id: str,
        model_zip_url: str,
        input: dict,
        outputs: dict,
        model_type: str,
        description: str = "",
        param_specs: List[dict] = None,
    ):
        """Doing 2 requests for initializing and creating version for a new trained model to the Clarifai platform.

        Args:
            model_id (str): Clarifai model id
            model_zip_url (str): url of zip of model
            input (dict): a dict where the key is clarifai input field and the value is triton model input,
                {clarifai_input_field: triton_input_filed}
            outputs (dict): a dict where the keys are clarifai output fields and the values are triton model outputs,
                {clarifai_output_field1: triton_output_filed1, clarifai_output_field2: triton_output_filed2,...}
            model_type (str): Clarifai model type.
            description (str, optional): a description of the model. Defaults to "".
            param_specs (List[dict]): list of dicts - keys are path, field_type, default_value, description. Default is None

        Returns:
            dict: Clarifai api response
        """
        init_resp = self.init_model(model_id, model_type, description)
        if init_resp["status"]["code"] != "SUCCESS":
            return init_resp
        version_resp = self.post_model_version(
            model_id, model_zip_url, input, outputs, param_specs
        )

        return version_resp

    def delete_model(self, model_id: str):
        """Delete model api by model id

        Args:
            model_id (str): Clarifai model id

        Returns:
            dict: clarifai api response
        """
        user_data_object = self.auth.get_user_app_id_proto()
        delete_model_response = self.stub.DeleteModel(
            service_pb2.DeleteModelRequest(
                user_app_id=user_data_object,
                model_id=model_id,
            ),
            metadata=self.auth.metadata,
        )

        return MessageToDict(delete_model_response, preserving_proto_field_name=True)

    def delete_model_version(self, model_id: str, version_id: str):
        """Delete specific version of model

        Args:
            model_id (str): Clarifai model id
            version_id (str): version id of model that will be removed

        Returns:
            dict: Clarifai API response
        """
        user_data_object = self.auth.get_user_app_id_proto()
        delete_model_response = self.stub.DeleteModelVersion(
            service_pb2.DeleteModelVersionRequest(
                user_app_id=user_data_object, model_id=model_id, version_id=version_id
            ),
            metadata=self.auth.metadata,
        )

        return MessageToDict(delete_model_response, preserving_proto_field_name=True)

    def get_model(self, model_id: str):
        """Get model by id

        Args:
            model_id (str): Clarifai model id

        Returns:
            dict: Clarifai API response
        """
        user_data_object = self.auth.get_user_app_id_proto()
        response = self.stub.GetModel(
            service_pb2.GetModelRequest(
                user_app_id=user_data_object,
                model_id=model_id,
            ),
            metadata=self.auth.metadata,
        )

        return MessageToDict(response, preserving_proto_field_name=True)
